''' Represents a sub or a master calibration image; handles all 
    FITs reading/writing and metadata from FITs header and/or filename parsing
'''

import os
import time
import glob
import numpy as np
from datetime import datetime
from astropy.io import fits
from kivy.logger import Logger

def is_fit(f):
    return f.lower().endswith('.fit') or f.lower().endswith('.fits')

class ImageNotReadyException(Exception):
    pass

def fits_in_dir(path):
    # get all fits in path, returning sorted by modification time
    fits = [f for f in glob.glob(os.path.join(path, '*')) if is_fit(f)]
    fits = {f: os.path.getmtime(f) for f in fits}
    return sorted(fits, key=fits.get)

def save_image(data=None, path=None, exposure=None, filt='L', sub_type='light', 
        nsubs=None, temperature=None):
    # saves sub or master
    try:
        hdu = fits.PrimaryHDU(data)
        hdr = hdu.header
        hdr['CREATOR'] = ('Jocular', 'v0.4')
        if exposure is not None:
            hdr['EXPOSURE'] = (exposure, 'seconds')
        if sub_type in {'master dark', 'master bias'}:
            hdr['FILTER'] = 'dark'
        else:
            hdr['FILTER'] = filt
        hdr['SUBTYPE'] = sub_type
        if temperature is not None:
            hdr['TEMPERAT'] = temperature
        if nsubs is not None:
            hdr['NSUBS'] = nsubs 
        hdu.writeto(path, overwrite=True)
        Logger.info('Image: saved to {:}'.format(path))
        for k, v in hdr.items():
            Logger.info('Image: {:9s} = {:}'.format(k, v))

    except Exception as e:
        Logger.warn('Image: unable to save image to {:} ({:})'.format(path, e))

class Image:

    # map from various FITS header names to Image attribute names
    hmap = {
        'exposure': 'exposure', 'exptime': 'exposure', 'expo': 'exposure', 'exp': 'exposure',
        'filter': 'filter', 'filt': 'filter',
        'subtype': 'sub_type', 'sub_type': 'sub_type', 'imagetyp': 'sub_type',
        'temperat': 'temperature', 'temp': 'temperature', 'ccd-temp': 'temperature',
        'nsubs': 'nsubs', 'stackcnt': 'nsubs'}

    # map from possible found filter names to the Jocular scheme
    filter_map = {'r': 'R', 'g': 'G', 'b': 'B', 'red': 'R', 'green': 'G', 'blue': 'B', 'dark': 'dark',  
        'ha': 'Ha', 'halpha': 'Ha', 'sii': 'SII', 'oiii': 'OIII', 'spect': 'spec', 'l': 'L', 'lum': 'L'}


    def __init__(self, path=None, verbose=False, check_image_data=False):
        ''' Generate Image instance from path. Throw ImageNotReadyException in cases where re-reading
            on next event cycle might be successful.
        '''

        if path is None:
            return

        if not is_fit(path):
            raise Exception('not a fits file {:}'.format(path))

        try:
            with fits.open(path) as hdu1:
                hdu1.verify('silentfix')
                hdr = hdu1[0].header
                # we check that the image is readable if necessary
                if check_image_data:
                    self.image = hdu1[0].data
                else:
                    self.image = None
        except Exception:
            self.image = None
            raise ImageNotReadyException('Cannot read fits header for {:}'.format(path))

        if int(hdr['NAXIS']) != 2:
            raise Exception('FITS does not contain 2D data')

        # extract any relevant information from FITs header

        fits_props = {}
        for k, v in hdr.items():
            if k.lower() in self.hmap:
                fits_props[self.hmap[k.lower()]] = v

        self.created_by_jocular = 'CREATOR' in hdr and hdr['CREATOR'].startswith('Jocular')

        self.shape = hdr['NAXIS1'], hdr['NAXIS2']
        self.shape_str = '{:}x{:}'.format(self.shape[0], self.shape[1])

        self.name = os.path.basename(path)
        self.path = path

        self.create_time = datetime.fromtimestamp(os.path.getmtime(path))
        self.arrival_time = datetime.now()
        self.age = (self.arrival_time - self.create_time).days

        # get info from filename
        name = os.path.splitext(self.name.lower())[0]

        # is it a master? check FITs subtype or name or nsubs
        master1 = fits_props.get('sub_type', '').lower().startswith('master')
        master2 = name.startswith('master')
        master3 = fits_props.get('nsubs', 0) > 1
        self.is_master = master1 or master2 or master3

        # is it a master? check FITs subtype or name
        # self.is_master = fits_props.get('sub_type', '').lower().startswith('master') or name.startswith('master')
        #self.is_master = name.startswith('masterflat') or name.startswith('masterdark') or name.startswith('masterbias')
        
        filename_props = self.parse_master(name[6:]) if self.is_master else self.parse_sub(name)

        # fits properties override filename properties
        props = {**filename_props, **fits_props}


        #----------- validate and normalise property values

        # sub type
        st = props.get('sub_type', '').lower()
        if st.startswith('master'):  # change 'master ' to 'master'
            st = st[6:].strip()
        if st[:4] in {'bias', 'dark', 'flat'}:
            props['sub_type'] = st[:4]
        else:
            props['sub_type'] = 'light'

        # exposure can be a number, or can a string be expressed in s or ms
        if 'exposure' in props and type(props['exposure']) == str:
            v = props['exposure'].lower()
            try:
                if v.endswith('ms'):
                    props['exposure'] = float(v[:-2]) / 1000
                elif v.endswith('s'):
                    props['exposure'] = float(v[-1])
                else:
                    props['exposure'] = float(v)
            except:
                del props['exposure']  # invalid exposure so remove it

        # filter might be R, r, red, filter: red, red filter, ...
        if 'filter' in props:
            v = props['filter'].lower()
            if v.endswith('filter'):
                v = v[:-6].strip()
            if v.startswith('filter'):
                v = v[6:].strip()
            if v in self.filter_map:
                props['filter'] = self.filter_map[v]
            else:
                del props['filter']

        # temperature format might be a number, or followed by C
        if 'temperature' in props and type(props['temperature']) == str:
            v = props['temperature'].lower()
            if v.endswith('c'):
                v = v[:-1].strip()
            try:
                props['temperature'] = float(v)
            except:
                del props['temperature']

        # number of subs
        if 'nsubs' in props and type(props['nsubs']) == str:
            v = props['nsubs']
            try:
                props['nsubs'] = int(v)
            except:
                del props['nsubs']

        # now assign relevant properties (and some defaults) to Sub
        self.exposure = props.get('exposure', None)

        # record the fact that exposure is purely an estimate (will allow user to later change it)
        self.exposure_is_an_estimate = self.exposure == None
        # record whether exposure has been estimated by user (using capture GUI)
        self.exposure_is_manual_estimate = False

        self.filter = props.get('filter', 'L')
        self.temperature = props.get('temperature', None)
        self.nsubs = props.get('nsubs', None)
        self.sub_type = props['sub_type']

        self.status = 'select'
        self.arrival_time = int(time.time())
        self.keyframe = False
        self.calibrations = {'dark': False, 'flat': False, 'bias': False}

        # force reload; is this necessary?
        # self.image = None

        # if verbose:
        #     self.describe()
        self.describe()

    def describe(self):
        # for debugging
        if self.is_master:
            Logger.info('Image: Master {:}'.format(self.name))
        else:
            Logger.info('Image: Sub {:}'.format(self.name))
        im = self.get_image()
        Logger.info('Image:    {:9s} {:14s} ({:} days) {:5s} {:}s {:} {:} {:} {:}'.format(
            self.shape_str, 
            self.create_time.strftime('%d %b %y %H:%M'),
            self.age,
            self.sub_type,
            self.exposure if self.exposure is not None else '?',
            '(est.)' if self.exposure_is_an_estimate else '',
            self.filter + ' filter' if self.filter else '',
            str(self.temperature) + 'C' if self.temperature is not None else '',
            str(self.nsubs) + ' subs in master' if self.nsubs is not None else ''
            ))
        Logger.info('Image:    min {:.4f} max {:.4f} mean {:.4f}'.format(
            np.min(im), np.max(im), np.mean(im)))


    def parse_sub(self, nm):
        # should extend this to support same parsing as calibration masters

        props = {}

        # pre v3 Jocular convention first
        try:
            [sub_type, filt, expo, subnum] = nm.split('_')
            props = {'sub_type': sub_type, 'filter': filt, 'exposure': int(expo[:-2]) / 1000.0}
        except:
            # parse name in one of 2 SLL formats
            nm = nm.lower()
            if nm.startswith('image_sll.'):
                nm = nm[10:]
            elif nm.startswith('image__'):
                nm = nm[7:]
            elif nm.startswith('image_'):
                nm = nm[6:]

            if nm.startswith('red') or nm.endswith('_red'):
                props['filter'] = 'R'
            elif nm.startswith('green') or nm.endswith('_green'):
                props['filter'] = 'G'
            elif nm.startswith('blue') or nm.endswith('_blue'):
                props['filter'] = 'B'
            elif nm.startswith('ha') or nm.endswith('_ha'):
                props['filter'] = 'Ha'
            elif nm[:4] in ['flat', 'dark', 'bias']:
                props['sub_type'] = nm[:4]

        return props

    def parse_master(self, nm):
        # nm  of form <type>_<key>=<value> where keys can come in any order
        # and keys are any of the things in the keymap

        prop_sep = '_'
        key_value_sep = '='
        parts = nm.split(prop_sep)
        props = {'sub_type': parts[0]}
 
        for p in parts[1:]:
            try:
                k, v = p.split(key_value_sep)
                if k in self.hmap:
                    props[self.hmap[k]] = v 
            except:
                pass

        return props

    def get_image(self):
        if self.image is None:
            try:
                with fits.open(self.path) as hdu:
                    bp = hdu[0].header['BITPIX']
                    if bp < 0:
                        self.image = hdu[0].data
                    else:
                        self.image = hdu[0].data / 2 ** bp
            except Exception as e:
                Logger.error('Image: cannot read image data from {:} ({:})'.format(self.path, e))
        return self.image


if __name__ == '__main__':

    fitsfiles = fits_in_dir('.')
    print('found {:} FITs'.format(len(fitsfiles)))
    for f in fitsfiles:
        print('-------------')
        try:
            s = Image(os.path.join('.', f), verbose=True)
            s.describe()
        except Exception as e:
            print('  Exception: {:}'.format(e))



''' Star extraction, centroid estimation and alignment.
'''

import warnings
import numpy as np

from scipy import signal
from skimage.measure import ransac
from skimage.transform import EuclideanTransform, matrix_transform, warp
from skimage.feature import blob_dog

from kivy.properties import ConfigParserProperty
from kivy.logger import Logger

from jocular.component import Component
from jocular.gradient import estimate_background, estimate_gradient


class Aligner(Component):

    do_align = ConfigParserProperty(1, 'Aligner', 'do_align', 'app', val_type=int)
    smooth_edges = ConfigParserProperty(1, 'Aligner', 'smooth_edges', 'app', val_type=int)
    ideal_star_count = ConfigParserProperty(30, 'Aligner', 'ideal_star_count', 'app', val_type=int)

    def __init__(self):
        super().__init__()
        self.min_sigma = 3
        self.max_sigma = 5
        self.min_stars = 5
        self.centroid_radius = 8
        r = self.centroid_radius
        patch_size = 2 * r + 1
        self.grid_x, self.grid_y = np.meshgrid(np.arange(patch_size), np.arange(patch_size))
        self.mask = np.sqrt((r - self.grid_x)**2 + (r - self.grid_y)**2) <= r
        self.star_intensity = None

    def on_new_object(self):
        self.reset()

    def reset(self):
        self.keystars = None
        self.warp_model = None
        self.align_count = 0
        self.starcounts = []
        self.info('reset')

    def update_status(self):
        sc = np.array(self.starcounts)
        self.info('{:}/{:} | stars {:}-{:}, mu:{:3.0f}'.format(
            self.align_count, len(sc), np.min(sc), np.max(sc), np.mean(sc)))
    
    def align(self, sub, centroids):
        # Align sub to current keystars, updating its status.
        
        min_inliers = 4

        # choose warp model to project existing keystars so they lie closer to centroids (in theory)
        keys = self.keystars
        if self.warp_model:
            keys = matrix_transform(self.keystars, self.warp_model.params)
        else:
            keys = self.keystars

        # find closest matching star to each transformed keystar
        nstars = min(len(keys), len(centroids))

        matched_stars = np.zeros((nstars, 2))
        for i, (x1, y1) in enumerate(keys):
            if i < nstars:
                matched_stars[i, :] = centroids[np.argmin([(x1 - x2) ** 2 + (y1 - y2) ** 2 for x2, y2 in centroids])]

        # do we have enough matched stars?
        if len(matched_stars) < self.min_stars:
            sub.status = 'nalign'
        else:            

            # apply RANSAC to find which matching pairs best fitting Euclidean model
            # can throw a warning in cases where no inliers (bug surely) which we ignore
            with warnings.catch_warnings():
                warnings.simplefilter('ignore')
                warp_model, inliers = ransac(
                    (np.array(self.keystars[:nstars]), matched_stars),
                    EuclideanTransform, 4, .5, max_trials=100)

            # managed to align
            if (inliers is not None) and (sum(inliers) >= min_inliers):
                # update warp model
                self.warp_model = warp_model

                # new in v0.4: can sometimes get artefacts that screw up platesolving
                # this is one way to handle them, but it isn't to all tastes and is slow
                # so allow use not to do it

                if self.smooth_edges:
                    # apply warp, ensuring values outside the range are filled with 
                    # a known value to mitigate edge effects on platesolving
                    sub.image = warp(sub.image, self.warp_model, order=3, 
                        preserve_range=True, mode='constant', cval=0.0001)

                    # identify unwarped points
                    unwarped = sub.image < .001

                    # grow these points inwards by convolution with block
                    # 76 ms to convolve is quite slow
                    unwarped = signal.convolve2d(unwarped, np.ones((9, 9)), mode='same')

                    # find positive points, but none away from boundary
                    pos = unwarped > 0
                    npix = 100
                    pos[npix:-npix, npix:-npix] = 0

                    # fill these with noise: ~ 14ms
                    bg_mu, bg_std = estimate_background(sub.image)
                    grad = estimate_gradient(sub.image)
                    # 10 ms
                    noi = np.random.normal(bg_mu, bg_std, sub.image.shape) * grad / np.mean(grad)
                    sub.image[pos] = noi[pos]

                else:
                    sub.image = warp(sub.image, self.warp_model, order=3, preserve_range=True)


                self.align_count += 1

                # change to select if it was nalign before; if reject before, leave it as reject
                if sub.status == 'nalign':
                    sub.status = 'select'
            else:
                sub.status = 'nalign'

        self.update_status()


    def extract_stars(self, im):
        # Extracts star coordinates. Return array of x, y coordinates.

        # if this is the first sub in the stack (perhaps after shuffle), find best intensity threshold
        if self.keystars is None:
            self.star_intensity = self.find_intensity_threshold(im)

        # extract using this threshold
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            stars = blob_dog(im, 
                min_sigma=self.min_sigma, 
                max_sigma=self.max_sigma, 
                threshold=self.star_intensity, 
                overlap=0)[:, [1, 0]]

        # new for v2: order stars by decreasing intensity
        intens = [im[int(y), int(x)] for x, y in stars]
        return np.array(stars)[[i for (v, i) in sorted((v, i) for (i, v) in enumerate(intens))][::-1]]


    def find_intensity_threshold(self, im):
        # Binary search to find intensity threshold for star extraction producing ideal star count 

        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            # image_cube = compute_image_cube(im, min_sigma=self.min_sigma, max_sigma=self.max_sigma)
            t0 , _ = estimate_background(im)
            # sometimes can fail due to background estimate being zero or too large
            if t0 < .001:
                t0 = .1
            if t0 > .5:
                t0 = .45
            crit = 20 / 100  # stop when within 20%
            lo, hi = 0, 1
            est = t0
            its = 0
            maxits = 8
            while True and (its < maxits):
                its += 1

                # this approach appears to cause problems on Windows, so going back to the slower approach
                # nstars = len(peak_local_max(image_cube, threshold_abs=est, footprint=np.ones((3,) * 3),
                #     threshold_rel=0.0, exclude_border=(0,) * 3))

                nstars = len(blob_dog(im, 
                    min_sigma=self.min_sigma, 
                    max_sigma=self.max_sigma, 
                    threshold=est, 
                    overlap=0)[:, [1, 0]])

                # Logger.debug('Aligner: iteration {:} nstars (slow method) {:}'.format(its, nstars))
                # stop when within crit percent of required
                if abs(nstars - self.ideal_star_count) / self.ideal_star_count < crit:
                    break
                if nstars < self.ideal_star_count:
                    hi = est
                else:
                    lo = est
                est = (hi + lo) / 2

        Logger.info('Aligner: bkgd {:6.4f} threshold {:6.4f} after {:} iterations'.format(t0, est, its))
        return est


    def compute_centroids(self, im, stars):
        # extract info from all candidate stars and return centroids
 
        r = self.centroid_radius
        rows, cols = im.shape
        star_data = np.zeros((len(stars), 2))
        ptr = 0
        for x, y in stars:
            x, y = int(x), int(y)
            if (x >= r) and (y >= r) and (x < (cols - r)) and (y < (rows - r)):
                imfrag = im[(y - r):(y + r + 1), (x - r):(x + r + 1)].copy()
                cx, cy = self.compute_centroid(imfrag)
                if cx > 0:
                    star_data[ptr, :] = [cx + x - r, cy + y - r]
                    ptr += 1

        return star_data[:ptr, :]

    def compute_centroid(self, imfrag):
        # slightly more sophisticated approach
        
        # mean intensity within masked circle
        mean_intensity = np.mean(imfrag[self.mask])
        
        # star pixels are in mask and above mean intensity
        candidates = self.mask & (imfrag > mean_intensity)
        
        if candidates.any():
            # estimate background and subtract from image
            mean_background = np.mean(imfrag[np.logical_not(candidates)])
            imfrag_sub = candidates * (imfrag - mean_background)       
            wim = np.sum(imfrag_sub)
            cx = np.sum(self.grid_x * imfrag_sub) / wim
            cy = np.sum(self.grid_y * imfrag_sub) / wim
            return cx, cy

        return -1, -1

    def process(self, sub):

        if not self.do_align:
            return

        # extract stars & compute centroids before aligning, if possible
        im = sub.get_image()
        raw_stars = self.extract_stars(im)
        centroids = self.compute_centroids(im, raw_stars)
        self.starcounts += [len(centroids)]

        if len(centroids) == 0:
            # no stars so don't do anything
            pass
        elif self.keystars is None:
            # first sub with stars so set keystars
            self.keystars = centroids
            self.align_count = 1
        else:
            self.align(sub, centroids)
            



''' Luminance channel manipulations, including B & W, stretch etc
'''

import numpy as np

from kivy.app import App
from kivy.uix.label import Label
from kivy.uix.boxlayout import BoxLayout
from kivy.metrics import dp
from kivy.properties import BooleanProperty, NumericProperty, OptionProperty
from kivy.core.window import Window

from jocular.stretch import stretch
from jocular.gradient import estimate_gradient, estimate_background, image_stats
from jocular.component import Component
from jocular.metrics import Metrics


class Monochrome(Component):

    redrawing = BooleanProperty(False)

    stretch = OptionProperty(
        "asinh", options=["linear", "log", "asinh", "gamma", "hyper"]
    )
    gradient = NumericProperty(100)
    white = NumericProperty(0.9)
    black = NumericProperty(0.1)
    p1 = NumericProperty(0.65)
    lift = NumericProperty(0)
    noise_reduction = NumericProperty(0)
    fine = NumericProperty(0)
    autoblack = BooleanProperty(True)
    show_image_stats = BooleanProperty(False)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.mono = None  # not 100% sure why this is needed on init
        self.view = Component.get("View")
        self.stacker = Component.get("Stacker")
        self.gui = App.get_running_app().gui

    def on_new_object(self):
        self.mono = None
        self.lum = None  #  luminosity (after applying stretch, B/W etc to mono)
        self._gradient = None  #  pixel-by-pixel zero-mean gradient estimate to subtract
        self._std_background = None  # std dev of background estimated from gradient
        self._blackpoint = None  # automatic estimate of blackpoint
        self._n = 0

    def on_p1(self, *args):
        self.adjust_lum()

    def on_white(self, *args):
        self.adjust_lum()

    def on_black(self, *args):
        if not self.autoblack:
            self.adjust_lum()

    def on_gradient(self, *args):
        self.adjust_lum()

    def on_lift(self, *args):
        self.adjust_lum()

    def on_noise_reduction(self, *args):
        self.adjust_lum()

    def on_stretch(self, *args):
        self.adjust_lum()

    def on_fine(self, *args):
        #  move black by a tiny amount
        if not self.autoblack:
            newblack = min(1, max(0, self.black + 0.1 * self.fine))
            self.gui.set("black", float(newblack))  # causes update

    def on_autoblack(self, *args):
        if self.autoblack:
            if hasattr(self, "mono") and self.mono is not None:
                self.update_blackpoint(self.mono)
                self.adjust_lum()

    def update_blackpoint(self, im):
        self._blackpoint, self._std_background = estimate_background(im)
        self.gui.set("black", float(self._blackpoint))

    def update_gradient(self, im):
        #  Estimate gradient and normalise to zero mean
        g = estimate_gradient(im)
        self._gradient = g - np.mean(g)

    def display_sub(self, im, do_gradient=False):
        """Called by Stacker when user selects sub, and by Capture, when displaying short subs
        Only compute gradient if light sub from stacker, not when calibration, nor when short
        """

        # ensure we are displaying subs
        if self.stacker.sub_or_stack == "stack":
            self.stacker.set_to_subs()

        # cache new image to allow user updates of B/W etc
        self.mono = im
        if do_gradient:
            self.update_gradient(im)
        if self.autoblack:
            self.update_blackpoint(im)
        self.view.display_image(self.luminance())

    def adjust_lum(self, *args):
        """User has changed control position so generate luminance and either
        display it (if sub or in mono mode) or advise multispectral of the update"""

        lum = self.luminance()
        if self.stacker.sub_or_stack == "sub":
            self.view.display_image(lum)
        else:
            if self.stacker.spectral_mode == "mono":
                self.view.display_image(lum)
            else:
                Component.get("MultiSpectral").luminance_updated(lum)

    def L_changed(self, L):
        if L is not None:
            lum = self.update_lum(L)
            self.view.display_image(lum)

    def update_lum(self, im):
        """Create gradient-adjusted luminance image; called here by L_changed and by
        MultiSpectral. Does not directly update display.
        """
        self.mono = im
        self.update_gradient(im)
        self.update_blackpoint(im)
        return self.luminance()

    def luminance(self, *args):
        # Applies black, white etc to current monochrome image, updating luminosity, returning the image.

        if self.mono is None:
            return

        im = self.mono



        # this is the point at which to compute image stats
        if self.show_image_stats:
            self.compute_image_stats()

        #  subtract some % of gradient if we have it computed (not the case for short subs)
        if (self._gradient is not None) and (self.gradient > 0.1):
            im = im - (self.gradient / 100) * self._gradient

        # set black based on automatic blackpoint estimate and lift setting
        #  we also allow lift settings in non-auto case

        if self.autoblack and (self._std_background is not None):
            black = max(0, self._blackpoint - self.lift * self._std_background)

        elif self._std_background is not None:
            black = max(0, self.black - self.lift * self._std_background)

        else:
            black = self.black

        im = (im - black) / (self.white - black)

        im[im > 1] = 1
        im[im < 0] = 0

        # timings: hyper:4, log: 7 asinh: 23, tanh: 6, gamma: 12
        im = stretch(
            im,
            method=self.stretch,
            param=self.p1,
            NR=self.noise_reduction,
            background=0
            if self._std_background is None
            else self.lift * self._std_background,
        )

        return im

    def compute_image_stats(self):
        if hasattr(self, "mono") and self.mono is not None:
            im = self.mono
            stats = image_stats(im)
            for s in [
                "min",
                "background",
                "100 x std. dev.",
                "central 75%",
                "99.9 percentile",
                "99.99 percentile",
                "max",
            ]:
                self.image_stats_fields[s].text = "{:}: {:.1f}%".format(
                    s, 100 * stats[s]
                )

    #  use kv for these?
    def make_image_stats_panel(self):
        p = BoxLayout(
            size_hint=(None, None), size=(dp(200), dp(200)), orientation="vertical"
        )
        App.get_running_app().gui.add_widget(p)
        self.image_stats_fields = {}
        for s in [
            "min",
            "background",
            "100 x std. dev.",
            "central 75%",
            "99.9 percentile",
            "99.99 percentile",
            "max",
        ]:
            self.image_stats_fields[s] = l = Label(
                text=s, size_hint=(None, None), size=(dp(200), dp(25))
            )
            p.add_widget(l)
        return p

    def on_show_image_stats(self, *args):
        if self.show_image_stats:
            if not hasattr(self, "image_stats"):
                self.image_stats = self.make_image_stats_panel()
            cx, cy = Metrics.get("origin")
            self.image_stats.center_x = cx
            self.image_stats.center_y = Window.height - dp(150)
            self.compute_image_stats()
        else:
            self.image_stats.y = 2 * Window.height



''' Handles new and previous object logic, and associated confirmation
    dialogues.
'''

import os
from datetime import date

from kivy.app import App
from kivy.logger import Logger
from kivy.properties import BooleanProperty, ConfigParserProperty
from kivy.lang import Builder
from kivy.uix.boxlayout import BoxLayout

from jocular.component import Component
from jocular.utils import add_if_not_exists, generate_observation_name, unique_member, move_to_dir
from jocular.widgets import JPopup

Builder.load_string('''

<MyLabel>:
    size_hint: 1, None
    height: dp(24)
    text_size: self.size
    markup: True
    halign: 'left'
 
<Confirmation>:
    orientation: 'vertical'
    size_hint: None, None
    width: dp(280)
    spacing: dp(0), dp(15)
    height:  _temp_box.height + _confirm_box.height + dp(20)

    BoxLayout:
        id: _temp_box
        size_hint: 1, None
        height: dp(60) if root.show_temperature else dp(0)
        opacity: 1 if root.show_temperature else 0
        disabled: not root.show_temperature
        orientation: 'vertical'

        MyLabel:
            text: 'Temperature {:}\N{DEGREE SIGN}C'.format(_temperature.value) if _temperature.value > -26 else 'Temperature not set'
        Slider:
            size_hint: 1, None
            height: dp(30)
            id: _temperature
            value: root.session.temperature if root.session.temperature is not None else -26
            on_value: root.session.temperature = self.value if self.value > -26 else None
            min: -26
            max: 40
            step: 1
            
    BoxLayout:
        id: _confirm_box
        size_hint: 1, None
        height: dp(40)
        opacity: 1

        Button:
            text: 'Save master' if root.calibration else 'Save'
            size_hint: .5, .8
            on_press: root.objectio.save_master() if root.calibration else root.objectio.save()
        Button:
            text: "Don't save master" if root.calibration else 'Cancel'
            size_hint: .5, .8
            on_press: root.objectio.save() if root.calibration else root.objectio.cancel_save()
''')


class Confirmation(BoxLayout):
    
    def __init__(self, objectio=None, **kwargs):
        self.objectio = objectio
        self.show_temperature = objectio.sub_type == 'dark'
        self.calibration = objectio.sub_type in {'dark', 'flat', 'bias'}
        self.session = Component.get('Session')
        super().__init__(**kwargs)


class ObjectIO(Component):

    existing_object = BooleanProperty(False)
    confirm_on_new = ConfigParserProperty(0, 'Confirmations', 'confirm_on_new', 'app', val_type=int)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.app = App.get_running_app()
        self.current_object_dir = None
        self.closing = False
        dformat = '%y_%m_%d'
        self.session_dir = os.path.join(self.app.get_path('captures'), date.today().strftime(dformat))
        add_if_not_exists(self.session_dir)

    def on_close(self):
        self.closing = True
        # self.confirm_new_object(closedown=True)

    def confirm_new_object(self, *args):
        # before saving, check for confirmations where required

        subs = Component.get('Stacker').subs

        # don't save if nothing changes, no current object, or no subs
        if not self.app.gui.something_has_changed or not self.current_object_dir or len(subs) == 0:
            Logger.debug('ObjectIO: no need to save anything')
            self.new_object()
            return

        # new in v0.4.5; if name has been changed, set as calibration type
        name = Component.get('DSO').Name
        if name.lower()[:4] in {'dark', 'flat', 'bias'}:
            self.sub_type = name.lower()[:4]
        else:
            self.sub_type = unique_member([s.sub_type for s in subs])

        # if calibration, ask user if they want to save master
        if self.sub_type in {'bias', 'flat', 'dark'}:
            self.popup = JPopup(title='Create new master{:}?'.format(self.sub_type), 
                content=Confirmation(objectio=self))
            self.popup.open()

        # user asked for confirmation?
        elif self.confirm_on_new:
            self.popup = JPopup(title='Confirm save (metadata changed)' if self.existing_object else 'Confirm save', 
                content=Confirmation(objectio=self))
            self.popup.open()

        # save silently
        else:
            self.save()


    def save(self, *args):
        # save info.json, handle rejects, and update observations table

        subs = Component.get('Stacker').subs
        Component.get('Metadata').set('exposure', unique_member([s.exposure for s in subs]))
        Component.get('Metadata').set('sub_type', self.sub_type)

        oldpath = self.current_object_dir
        Component.save_object()

        # check if name has been changed
        name = Component.get('Metadata').get('Name', default='')
        if name:
            session_dir, object_dir = os.path.split(self.current_object_dir)
            if name != object_dir:
                # user has changed name, so generate a (unique) new folder name
                new_name = generate_observation_name(session_dir, prefix=name)
                # change directory name; any problem, don't bother
                try:
                    os.rename(self.current_object_dir, 
                        os.path.join(session_dir, new_name)) 
                    self.current_object_dir = os.path.join(session_dir, new_name)
                    self.info('saved {:}'.format(new_name))
                except Exception as e:
                    Logger.exception('ObjectIO: cannot change name ({:})'.format(e))
                    self.warn('cannot change name')

        # save metadata, and if successful update observations
        newpath = self.current_object_dir
        try:
            Component.get('Metadata').save(newpath)            
            Component.get('Observations').update(oldpath, newpath)
            Component.get('ObservingList').new_observation()
        except Exception as e:
            Logger.exception('ObjectIO: OSError saving info3.json to {:} ({:})'.format(newpath, e))
            self.warn('saving metadata')

        self.new_object()
        if hasattr(self, 'popup'):
            self.popup.dismiss()

    def save_master(self):
        subs = Component.get('Stacker').subs
        Component.get('Calibrator').save_master(
            exposure=unique_member([s.exposure for s in subs]), 
            temperature=Component.get('Session').temperature,
            filt=unique_member([s.filter for s in subs]),
            sub_type=self.sub_type)
        self.save()

    def cancel_save(self, *args):
        self.popup.dismiss()

    def new_object(self, *args):
        if self.closing:
            return
        self.current_object_dir = None
        self.existing_object = False
        self.sub_type = None
        Component.get('Metadata').reset()
        Component.initialise_new_object()
        # App.get_running_app().gui.enable()
        # this should be done by the camera mode; so don't do it for watched or controlled watched
        self.app.gui.set('frame_script', True, update_property=True)

    def load_previous(self, path):
        # previous will have been saved by this point
        self.existing_object = True
        self.changed = False # we start off assuming nothing has changed
        Component.get('Metadata').load(path)
        gui = self.app.gui
        gui.enable()
        gui.set('show_reticle', False, update_property=True)
        self.current_object_dir = path
        Component.initialise_new_object()
        stacker = Component.get('Stacker')
        stacker.recompute()
        Component.get('Capture').disable_captures()

        # # disable change to exposure if we are sure it is not an estimate
        subs = stacker.subs
        if len(subs) == 0:
            return

        sub_type=unique_member([s.sub_type for s in subs])
        if subs[0].exposure_is_an_estimate:
            self.app.gui.enable(['exposure_button', '{:}_script'.format(sub_type)])

        # add info on exposure and filter(s)
        expo = unique_member([s.exposure for s in subs])
        if expo is None:
            expo = Component.get('Metadata').get('exposure')
        if expo is None:
            expo = 0

        Component.get('CaptureScript').set_external_details(exposure=expo, 
            sub_type=sub_type, 
            filt=''.join({s.filter for s in subs}))


    def new_sub_from_watcher(self, sub):
        ''' Receiving a new sub from watcher might be the first we hear of 
        a new object, so method checks if we have a current object directory
        '''

        stacker = Component.get('Stacker')
        sub_type = sub.sub_type
 
        if self.current_object_dir is None:
            # new object, so check if calibration or light
            if sub_type in {'dark', 'flat', 'bias'}:
                self.current_object_dir = os.path.join(self.session_dir, 
                    generate_observation_name(self.session_dir, prefix=sub_type))
                stacker.sub_type = sub.sub_type  # is this needed any more?
                Component.get('DSO').Name = sub_type
            else:
                self.current_object_dir = os.path.join(self.session_dir, 
                    generate_observation_name(self.session_dir, prefix=Component.get('DSO').Name))
                self.app.gui.disable(['dark_script', 'flat_script', 'load_previous'])
            add_if_not_exists(self.current_object_dir)

        newpath = os.path.join(self.current_object_dir, sub.name)

        try:
            os.rename(sub.path, newpath)
            sub.path = newpath
            stacker.add_sub(sub)
        except Exception as e:
            Logger.error('ObjectIO: cannot add sub to stack ({:})'.format(e))

        # if we have an estimated exposure for this sub and if so, allow user to change it via GUI
        # disable change to exposure if we are sure it is not an estimate

        if sub.exposure_is_an_estimate:
            # self.app.gui.disable(['capturing', 'filter_button', 'frame_script', 'seq_script',
            self.app.gui.disable(['filter_button', 'frame_script', 'seq_script',
            'light_script', 'bias_script', 'dark_script', 'flat_script'])
            self.app.gui.enable(['exposure_button', '{:}_script'.format(sub_type)])

            Component.get('CaptureScript').set_external_details(sub_type=sub_type)


    def save_original(self, path):
        ''' Move non-Jocular sub from path (in Watched) to current object subdirectory
        '''
        if self.current_object_dir is None:
            self.current_object_dir = os.path.join(self.session_dir, 
                generate_observation_name(self.session_dir, prefix=Component.get('DSO').Name))
            add_if_not_exists(self.current_object_dir)
        move_to_dir(path, os.path.join(self.current_object_dir, 'originals'))




''' Support for RA/Dec text representations and conversions.
'''

class RA(float):
    def __new__(cls, ra):
        if type(ra) == str:
            try:
                h, m ,s = ra.split()
                if h[-1] == 'h':
                    h = h[:-1]
                ra = float(h)*15 + float(m)/4 + float(s)/240
                if (ra < 0) | (ra > 360):
                    return float('nan')
            except:
                return float('nan')
        return float.__new__(cls, ra)
 
    def __str__(self):
        hr, rest = divmod(self, 15)
        minit, rest = divmod(rest*4, 1)
        return "{:02d}h {:02d} {:02.0f}".format(int(hr), int(minit), rest*60)

class Dec(float):
    def __new__(cls, dec):
        if type(dec) == str:
            try:
                d, m, s = dec.split()
                if d[-1] == '\u00b0':
                    d = d[:-1]
                if d[0] == '-':
                    dec = float(d) - float(m)/60 - float(s)/3600
                else:
                    dec = float(d) + float(m)/60 + float(s)/3600
                if (dec < -90) | (dec > 90):
                    return float('nan')
            except:
                return float('nan')
        return float.__new__(cls, dec)

    def __str__(self):
        if self >= 0:
            dd, rest = divmod(self, 1)
            dmin, rest = divmod(rest*60, 1)
            return "+{:d}\u00b0 {:02d} {:02.0f}".format(int(dd), int(dmin), rest*60)
        else:
            dd, rest = divmod(-self, 1)
            dmin, rest = divmod(rest*60, 1)
            return "-{:d}\u00b0 {:02d} {:02.0f}".format(int(dd), int(dmin), rest*60)




"""Hoverable Behaviour (changing when the mouse is on the widget by O. Poyen.
License: LGPL
"""
__author__ = 'Olivier POYEN'

from kivy.properties import BooleanProperty, ObjectProperty
from kivy.core.window import Window

class HoverBehavior(object):
    """Hover behavior.
    :Events:
        `on_enter`
            Fired when mouse enter the bbox of the widget.
        `on_leave`
            Fired when the mouse exit the widget 
    """

    hovered = BooleanProperty(False)
    border_point= ObjectProperty(None)
    '''Contains the last relevant point received by the Hoverable. This can
    be used in `on_enter` or `on_leave` in order to know where was dispatched the event.
    '''

    def __init__(self, **kwargs):
        self.register_event_type('on_enter')
        self.register_event_type('on_leave')
        Window.bind(mouse_pos=self.on_mouse_pos)
        super(HoverBehavior, self).__init__(**kwargs)

    def on_mouse_pos(self, *args):
        if not self.get_root_window():
            return # do proceed if I'm not displayed <=> If have no parent
        pos = args[1]
        #Next line to_widget allow to compensate for relative layout
        inside = self.collide_point(*self.to_widget(*pos))
        if self.hovered == inside:
            #We have already done what was needed
            return
        self.border_point = pos
        self.hovered = inside
        if inside:
            self.dispatch('on_enter')
        else:
            self.dispatch('on_leave')

    def on_enter(self):
        pass

    def on_leave(self):
        pass

from kivy.factory import Factory
Factory.register('HoverBehavior', HoverBehavior)




''' Serves stars and catalogues to other components, notably the
    platesolver, annotator, and observing list
'''

import os
import glob
import csv
import json
from copy import deepcopy
import numpy as np
from kivy.app import App
from kivy.logger import Logger
from kivy.properties import ConfigParserProperty

from jocular.component import Component

def intstep(d, step):
    return int(np.floor(step * np.floor(d / step)))

class Catalogues(Component):

    exclude_cats = ConfigParserProperty('VS', 'Catalogues', 'exclude_cats', 'app', val_type=str)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.app = App.get_running_app()
        self.star_db = None

        ''' load object types and normalise colour values to 0-1 
        '''
        with open(self.app.get_path('object_types.json'), 'r') as f:
            self.object_types = json.load(f)
        for v in self.object_types.values():
            v['col'] = [i/255. for i in v['col']]

        self.dsos = None
        self.on_new_object()

    def get_object_types(self):
        return self.object_types

    def on_new_object(self):
        ''' Every time we get a new object, check for the existence
            of the catalogues. This allows a user to load them without
            restarting Jocular
        '''
        if self.star_db is None:
            try:
                self.star_db  = np.load(self.app.get_path('star_db'))
            except Exception as e:
                pass

    def has_platesolving_db(self):
        return self.star_db is not None

    def in_tile(self, tile, ras, decs):
        ''' return indices of coordinates within tile
        '''
        rmin, rmax = tile['min_ra'], tile['max_ra']
        dmin, dmax = tile['min_dec'], tile['max_dec']
        if rmin < rmax:
            return (decs >= dmin) & (decs < dmax) & (ras >= rmin) & (ras < rmax)
        else:
            return (decs >= dmin) & (decs < dmax) & ((ras >= rmin) | (ras < rmax))

    def get_tiles(self, tile, rstep, dstep):
        ''' return database tiles covering region defined in tile
        '''
        rmin, rmax = tile['min_ra'], tile['max_ra']
        dmin, dmax = tile['min_dec'], tile['max_dec']

        dec_tiles = np.arange(intstep(dmin, dstep), intstep(dmax, dstep) + dstep, dstep)
        if rmin < rmax:
            ra_tiles = np.arange(intstep(rmin, rstep), intstep(rmax, rstep) + rstep, rstep)
        else:
            # handle wraparound
            ra_tiles = \
                list(range(intstep(rmin, rstep), 360, rstep)) + \
                list(range(0, intstep(rmax, rstep) + rstep, rstep))
        return ra_tiles, dec_tiles        

    def get_platesolving_stars(self, tile=None):
        # get stars in tile by reading npys that are spaced at 30 x 10 degrees in RA/Dec

        if self.star_db is None or tile is None:
            return

        ra_tiles, dec_tiles = self.get_tiles(tile, 30, 10)
        ras, decs, mags = np.array([]), np.array([]), np.array([])
        for ra in ra_tiles:
            for dec in dec_tiles:
                quad = 'r{:}_d{:}_'.format(ra, dec)
                ras = np.append(ras, self.star_db[quad+'ra'])
                decs = np.append(decs, self.star_db[quad+'dec'])
                mags = np.append(mags, self.star_db[quad+'mag']/100.) # mags are ints * 100

        locs = self.in_tile(tile, ras, decs)
        return ras[locs], decs[locs], mags[locs]

    def get_basic_dsos(self):
        if self.dsos is None:
            self.load_basic_dsos()
        return self.dsos

    def load_basic_dsos(self):
        ''' We load the shipped and user catalogues at the start once only
        '''

        ''' construct list of .csv files; we load shipped then user cats
            with later loads overwriting earlier loads of duplicate
            objects (thereby supporting user updates)
        '''

        shipped = glob.glob(os.path.join(self.app.get_path('dsos'), '*.csv'))
        usercats = glob.glob(os.path.join(self.app.get_path('catalogues'), '*.csv'))

        self.dsos = {}
        for f in shipped + usercats:
            user_objects = f.endswith('user_objects.csv')
            dupes = 0
            with open(f, newline='') as csvfile:
                reader = csv.reader(csvfile, delimiter=',')
                hdr = next(reader)
                if {'Name', 'RA', 'Dec', 'OT', 'Con'} <= set(hdr):
                    cnames = ['Name', 'RA', 'Dec', 'Con', 'OT', 'Mag', 'Diam', 'Other']
                    cols = {c: (hdr.index(c) if c in hdr else None) for c in cnames}
                    cols = {c: i for c, i in cols.items() if i is not None}
                    for row in reader:
                        try:
                            dd = {c: row[i] for c, i in cols.items()}
                            name = '{:}/{:}'.format(dd['Name'], dd['OT'])
                            if name in self.dsos and not user_objects:
                                dupes += 1
                            else:
                                self.dsos[name] = dd
                        except:
                            if len(row) > 0:
                                Logger.warn('Catalogues: error in catalogue {:}, {:}'
                                    .format(f, row))
                else:
                    Logger.warn('Catalogues: {:} no Name/RA/Dec/OT/Con'.format(f))
            Logger.info('Catalogues: loaded {:} ({:} duplicates)'.format(
                os.path.basename(f), dupes))

        # convert all numeric cols to floats
        for k, v in self.dsos.items():
            for col in ['RA', 'Dec', 'Mag', 'Diam']:
                if col in v:
                    v[col] = float(v[col] if v[col] else 'nan')

        Logger.info('Catalogues: loaded {:} DSOs from {:} catalogues'.format(
            len(self.dsos), len(shipped + usercats)))

    def get_annotation_dsos(self, tile=None):
        ''' Retrieve all DSOs in the specified tile. Use basic DSOs combined
            with any deeper DSOs available. Return as a dict with dso names as
            keys and values are dicts of object properties

            To do: streamline using in_tile etc so RA wraparound is handled in one place
        '''

        if tile is None:
            return {}

        dsos = self.get_basic_dsos()
        dmin, dmax = tile['min_dec'], tile['max_dec']
        in_dec = {k: v for k, v in dsos.items() 
            if (v['Dec'] >= dmin) & (v['Dec'] < dmax)}
        rmin, rmax = tile['min_ra'], tile['max_ra']
        if rmin < rmax:
            matches = {k: v for k, v in in_dec.items() 
                if (v['RA'] >= rmin) & (v['RA'] < rmax)}
        else:
            # handle RA wrap-around
            matches = {k: v for k, v in in_dec.items() 
                if (v['RA'] >= rmin) | (v['RA'] < rmax)}

        # load any deep catalogues that user has placed in catalogues
        excluded = self.exclude_cats.lower().split()
        deepcats = glob.glob(os.path.join(self.app.get_path('catalogues'), '*.npz'))
        for cat in deepcats:
            if os.path.basename(cat)[:-4].lower() not in excluded:
                matches = {**matches, **self.load_deepcat(cat, tile)}

        # return copy in case receiver decides to modify things
        return deepcopy(matches)

    def load_deepcat(self, nm, tile):

        cat = np.load(nm, allow_pickle=True)
        try:
            rstep, dstep = cat['ra_step'][0], cat['dec_step'][0]
        except:
            return {}

        ra_tiles, dec_tiles = self.get_tiles(tile, rstep, dstep)

        # extract objects from all necessary tiles
        cols = cat['columns']
        OT = cat['OT'][0]
        catname = cat['catname'][0]

        matches = {}
        for ra in ra_tiles:
            for dec in dec_tiles:
                subdat = cat['r{:}_d{:}'.format(ra, dec)][0]
                ras, decs = subdat['RA'], subdat['Dec']
                locs = self.in_tile(tile, ras, decs)
                for l in np.where(locs)[0]:
                    matches[subdat['Name'][l] + '/' + OT] = {c: subdat[c][l] for c in cols}

        # add object type and catalogue
        for k, v in matches.items():
            v['OT'] = OT
            v['Cat'] = catname

        return matches






''' Manages calibration, alignment and stacking of incoming subs
'''

import os
import glob
import numpy as np
from datetime import datetime

from kivy.app import App
from kivy.logger import Logger
from kivy.properties import (BooleanProperty, NumericProperty, OptionProperty, 
    StringProperty, ListProperty, ConfigParserProperty)
from kivy.clock import Clock

from jocular.component import Component
from jocular.utils import percentile_clip, s_to_minsec, move_to_dir, purify_name
from jocular.widgets import JBubble
from jocular.image import Image, fits_in_dir

def combine_stack(stk, method='mean'):
    if len(stk) == 1:
        s = stk[0]
    elif len(stk) == 2:
        s = np.mean(stk, axis=0)
    elif method == 'mean':
        s = np.mean(stk, axis=0)
    elif method == 'median':
        s = np.median(stk, axis=0)
    else: 
        s = percentile_clip(stk, perc=int(method))
    return s

class Stacker(Component):

    selected_sub = NumericProperty(-1) # index of currently selected sub starts at 0
    sub_or_stack = StringProperty('sub')
    L_plus_other = StringProperty('Ha')
    subs = ListProperty([])
    animating = BooleanProperty(False)
    combine = OptionProperty('mean', options=['mean', 'median', '70', '80', '90'])
    speed = NumericProperty(2)
    spectral_mode = OptionProperty('mono', options=['mono', 'LRGB', 'L+'])
    confirm_on_clear_stack = ConfigParserProperty(0, 'Confirmations', 'confirm_on_clear_stack', 'app', val_type=int)
    reload_rejected = ConfigParserProperty(0, 'rejects', 'reload_rejected', 'app', val_type=int)
    use_TOA_for_exposure = ConfigParserProperty(1, 'Exposure', 'use_TOA_for_exposure', 'app', val_type=int)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.app = App.get_running_app()
        self.on_spectral_mode()

        self.sub_colors = {'select': (0, 1, 0, .6), 
            'reject': (.5, .5, .5, .6), 
            'nalign': (1, 0, 0, .6),
            'delete': (.1, .1, .1, .6)}

    def on_new_object(self):
        ''' Prepare for existing or new object. For existing object, loading takes
        place after all components' on_new_object methods have been called 
        '''

        self.reset()
        self.orig_rejects = set({})
        self.original_exposure = 0
        settings = Component.get('Metadata').get({'sub_type', 'rejected', 'exposure'})
        sub_type = settings.get('sub_type', 'light')
        cod = Component.get('ObjectIO').current_object_dir

        # in pre v3 we used a 'rejected' list to store names of rejects; now we don't
        rejected = set(settings.get('rejected', []))
        exposure = settings.get('exposure', None)
        if cod is not None:
            # move any in rejects if user has requested this
            if self.reload_rejected:
                for f in fits_in_dir(os.path.join(cod, 'rejects')):
                    try:
                        bn = os.path.basename(f)
                        os.rename(f, os.path.join(cod, bn))
                        rejected.add(bn)
                    except Exception as e:
                        Logger.warn('Stacker: problem moving rejected on load {:} ({:})'.format(f, e)) 
            # now load all (which now will include prev rejects)
            for f in fits_in_dir(cod):
                try:
                    im = Image(f)
                    im.status = 'reject' if os.path.basename(f) in rejected else 'select'
                    if im.exposure is None:
                        im.exposure = exposure
                    if im.sub_type is None:
                        im.sub_type=sub_type
                    self.subs.append(im)
                except Exception as e:
                    Logger.warn('Stacker: problem reading image {:} ({:})'.format(f, e))
            # if we have loaded up some calibration subs, ensure that we mark things as changed
            if not self.is_empty():
                self.changed = self.subs[0].sub_type in {'dark', 'bias', 'flat'}
                self.original_exposure = self.subs[0].exposure
            # store rejected to see if anything has changed and needs saving
            self.orig_rejects = rejected
 
    def reset(self):
        # Called when we have a new object and when user clears stack
        self.stack_cache = {}
        self.subs.clear()
        self.selected_sub = -1
        self.update_stack_scroller()
        self.set_to_subs()
        Component.get('View').reset()  # might not be needed as View also does a reset!
        self.info('reset stack')

    def on_save_object(self):
        # move rejected subs to 'rejected'; we won't touch non-aligned as they are salvageable often
        cod = Component.get('ObjectIO').current_object_dir
        for s in self.subs:
            if s.status == 'reject':
                move_to_dir(os.path.join(cod, s.name), 'rejects')

    def is_empty(self):
        return len(self.subs) == 0

    def update_status(self):
        d = self.describe()
        if d:
            self.info('{:} | {:}x{:.0f}s | {:}'.format(
                s_to_minsec(d['total_exposure']), 
                d['nsubs'], 
                d['sub_exposure'], 
                d['filters']))

    def describe(self):
        # return info useful to snapshotter (for the moment)

        if len(self.subs) > 0 and self.sub_or_stack == 'sub':
            s = self.subs[self.selected_sub]
            return {
                'nsubs': 1,
                'total_exposure': s.exposure if s.exposure else 0,
                'sub_exposure': s.exposure if s.exposure else 0,
                'multiple_exposures': False,
                'dark': 'dark' in s.calibrations,
                'flat': 'flat' in s.calibrations,
                'bias': 'bias' in s.calibrations,
                'filters': 'filter: {:}'.format(s.filter)}

        subs = [s for s in self.subs[:self.selected_sub+1] if s.status == 'select']
        if len(subs) == 0:
            return None

        expos = [s.exposure if s.exposure else 0 for s in subs]
        filts = [s.filter for s in subs]
        fstr = ''
        for f in ['L', 'R', 'G',  'B', 'Ha', 'OIII', 'SII']:
            if filts.count(f) > 0:
                fstr += '{:}{:} '.format(filts.count(f), f)

        return {
            'nsubs': len(subs),
            'total_exposure': sum(expos),
            'sub_exposure': max(set(expos), key=expos.count),
            'multiple_exposures': len(set(expos)) > 1,
            'dark': 'dark' in subs[0].calibrations,
            'flat': 'flat' in subs[0].calibrations,
            'bias': 'bias' in subs[0].calibrations,
            'filters': fstr.strip()}

    def get_filters(self):
        if self.is_empty():
            return None
        return {s.filter for s in self.subs[:self.selected_sub+1] if s.status == 'select'}

    def clear(self, *args):
        ''' Clear existing stack by moving/deleting FITs and resetting
        '''

        if len(self.subs) == 0:
            return

        if self.confirm_on_clear_stack:
            JBubble(actions={'Really clear?': self._clear}, loc='mouse').open()
        else:
            self._clear()

    def _clear(self, *args):
        # clear stack auxilliary method since might come after confirmation

        cod = Component.get('ObjectIO').current_object_dir
        for nm in glob.glob(os.path.join(cod, '*.fit*')):
            move_to_dir(os.path.join(cod, nm), 'deleted')

        # perform necessary resets
        Component.get('Aligner').reset()
        self.reset()
        Component.get('Capture').soft_reset()
        self.changed = False # indicate nothing has changed

    def delete_sub(self, *args):
        if self.is_empty():
            return
        sub_num = self.selected_sub
        if sub_num == 0:
            return
        cod = Component.get('ObjectIO').current_object_dir
        move_to_dir(os.path.join(cod, self.subs[sub_num].name), 'deleted')
        del self.subs[sub_num]
        self.selected_sub -= 1

    def on_animating(self, *args):
        if self.is_empty():
             return
        if self.animating:
            self.play_event = Clock.schedule_interval(self.increment_sub, 1/self.speed)
        else:
            if hasattr(self, 'play_event'):
                Clock.unschedule(self.play_event)

    def on_sub_or_stack(self, *args):
        self.on_selected_sub()

    def on_L_plus_other(self, *args):
        self.stack_changed()

    def on_speed(self, *args):
        if self.animating:
            Clock.unschedule(self.play_event)
            self.play_event = Clock.schedule_interval(self.increment_sub, 1/self.speed)

    def first_sub(self, *args):
        self.selected_sub = 0

    def last_sub(self, *args):
        self.selected_sub = len(self.subs) - 1

    def previous_sub(self, *args):
        if self.selected_sub > 0:
            self.selected_sub -= 1

    def next_sub(self, *args):
        if self.selected_sub < len(self.subs) - 1:
            self.selected_sub += 1

    def on_combine(self, *args):
        self.stack_changed()

    def set_to_subs(self):
        self.app.gui.set('sub_or_stack', 'sub', update_property=True)

    def toggle_selected_sub(self, *args):
        # user selects displayed sub icon to toggle its selected status

        if self.is_empty():
            return

        current_sub = self.selected_sub
        sub = self.subs[current_sub]
        if sub.status == 'select':
            sub.status = 'reject'
        elif sub.status == 'reject':
            sub.status = 'select'

        self.update_stack_scroller()    # updates colours
        self.stack_changed()            # will check if sub or stack

        # check if anything has changed
        self.check_for_changes()

    def on_spectral_mode(self, *args):
        self.stack_changed()
 
    def stack_changed(self):
        ''' Called whenever we might need to update the stack e.g. sub added/selected
        deselected/change in stack view/change in mode
        '''

        if self.is_empty():
            return
        if self.sub_or_stack == 'sub':
            return

        # see if we can satisfy user's non-mono preferences and if not, drop thru to mono
        filters = self.get_filters()

        if self.spectral_mode == 'LRGB':
            Component.get('MultiSpectral').LRGB_changed(
                L=self.get_stack('L'), 
                R=self.get_stack('R'), 
                G=self.get_stack('G'), 
                B=self.get_stack('B'))

        elif self.spectral_mode == 'L+':
            Component.get('MultiSpectral').L_plus_changed(
                L=self.get_stack('L'), 
                layer=self.get_stack(self.L_plus_other)) 

        else:
            Component.get('Monochrome').L_changed(self.get_stack(filt='all'))
            Component.get('MultiSpectral').reset()            

        self.update_status()

 
    def update_stack_scroller(self, *args):
        # sub_labels is array representing the stack
        # these are sub positions on the screen

        self.update_status()
        gui = self.app.gui.gui
        labs = [gui[n]['widget'] for n in ['sub_m2', 'sub_m1', 'sub_0', 'sub_p1', 'sub_p2']]

        # set background to transparent and text to blank to start with
        for l in labs:
            l.text = ''
            l.background_color[-1] = 0

        # we have some subs
        fw = Component.get('FilterWheel')
        for i, l in enumerate(labs, start=self.selected_sub - 2):
            # if position is occupied by a sub
            if (i >= 0) and (i < len(self.subs)):
                l.text = str(i + 1)
                l.color = self.sub_colors[self.subs[i].status]
                l.background_color = fw.filter_properties[self.subs[i].filter]['bg_color']

    def increment_sub(self, dt):
        if self.selected_sub == len(self.subs) - 1:
            self.selected_sub = 0
        else:
            self.selected_sub += 1


    def get_screen(self, dt=None):
        im = Component.get('Snapshotter').snap(return_image=True).copy()
        self.ims.append(im.convert('L'))
        if len(self.ims) < len(self.subs):
            self.selected_sub += 1
            Clock.schedule_once(self.get_screen, 0)
        else:
            save_path = '{:} {:}.gif'.format(
                    os.path.join(self.app.get_path('snapshots'), 
                        purify_name(Component.get('DSO').Name)), 
                    datetime.now().strftime('%d%b%y_%H_%M_%S'))
            self.ims[0].save(save_path, 
                save_all=True, 
                append_images=self.ims[1:], 
                optimize=True, 
                # include_color_table=False,
                duration=1000 / self.speed, 
                loop=0)
            Component.get('Snapshotter').info('saved animation')

    def make_animated_gif(self):
        self.ims = []
        self.selected_sub = 0
        Clock.schedule_once(self.get_screen, 0)

    def sub_added(self):
        # called when new sub added to the stack by Watcher
        if self.selected_sub == len(self.subs) - 2:  # was on last sub before new one came in
            self.selected_sub = len(self.subs) - 1
        else:
            self.update_stack_scroller()

    def on_selected_sub(self, *args):
        # This is the key method that dictates whether the display has changed
        # on new sub, change of sub selection etc and hence stack updates

        s = self.selected_sub
        if (s < 0) or (s >= len(self.subs)):
            return

        if self.sub_or_stack == 'sub':
            ss = self.subs[s]
            Component.get('Monochrome').display_sub(ss.get_image(), 
                do_gradient=ss.sub_type=='light')
        else:
            self.stack_changed()

        self.update_stack_scroller()


    def get_image_for_platesolving(self):
        ''' Currently gets displayed image on stack, but what if we 
            are framing etc?
        '''
        try:
            if self.sub_or_stack == 'sub':
                im = self.subs[self.selected_sub].get_image()
            else:
                im = self.get_stack()
            return Component.get('View').do_flips(im)
        except Exception as e:
            Logger.warning('Stacker: no image for platesolving')
            return None

    def get_selected_sub_count(self):
        # used by calibrator
        return len([1 for s in self.subs if s.status == 'select'])

    def get_stack(self, filt='all', calibration=False):
        ''' Return stack of selected subs of the given filter if provided. Normally uses 
        those subs up to the currently selected sub on the display. Responsible for 
        caching results to prevent expensive recomputes. For calibration, uses all subs.
        '''

        # we have no subs
        if not self.subs:
            return None

        # subs up to current selected sub, except in case of calibration
        if calibration:
            subs = self.subs
        else:
            subs = self.subs[:self.selected_sub+1]

        # if no filter, choose all subs, otherwise restrict
        if filt == 'all':
            stk = [s for s in subs if (s.status=='select')]
        else:
            stk = [s for s in subs if (s.status=='select') & (s.filter==filt)]

        # none remain
        if not stk:
            return None

        # if calibrating, use user-selected method, otherwise use mean for non-L channels
        if calibration or filt in ['L', 'all']:
            method = self.combine
        else:
            method = 'mean'

        sub_names = set([s.name for s in stk])

        # check if we already have this in the cache
        if filt in self.stack_cache:
            cached = self.stack_cache[filt]
            if (cached['method'] == method) and (cached['sub_names'] == sub_names):
                return cached['stack']

        # if not, we need to update
        stacked = combine_stack(np.stack([s.get_image() for s in stk], axis=0), method=method)

        # cache results
        self.stack_cache[filt] = {
            'stack': stacked, 
            'method': method,
            'sub_names': sub_names
            }

        return stacked

    def add_sub(self, sub):
        # for sub coming from Watcher

        ''' if exposure is estimated, then find estimate. Rules, in priority order:
            (1) if the user has provided a manual estimate via GUI, then always use that.
            (2) if the user allows TOA differences and there are enough subs, use that.
                and propogate estimate to rest of stack so it is continually updated
            (3) otherwise for now use value for this sub type from GUI, even if user
                hasn't altered it; note that it is still not a manual estimate to
                give a chance for re-estimation using TOA later
        '''

        if sub.exposure_is_an_estimate:
            # if first sub used manual estimate, use that estimate
            if len(self.subs) > 0 and self.subs[0].exposure_is_manual_estimate:
                sub.exposure = self.subs[0].exposure
                sub.exposure_is_manual_estimate = True
            # if user wishes to to TOA for exposure, try to use that
            elif self.use_TOA_for_exposure and len(self.subs) > 3:
                dTOA = int(np.mean(np.diff([s.arrival_time for s in self.subs])))
                sub.exposure = dTOA
                # propagate it to all subs
                for s in self.subs:
                    s.exposure = dTOA
                    s.exposure_is_manual_estimate = False
            # otherwise use current exposure indicated for this sub type
            else:
                sub.exposure = Component.get('CaptureScript').get_current_exposure(sub.sub_type)
                sub.exposure_is_manual_estimate = False

            # tell interface
            Component.get('CaptureScript').set_external_details(sub_type=sub.sub_type, exposure=sub.exposure)


        self.process(sub)
        self.subs.append(sub)
        self.sub_added()

        self.changed = not self.is_empty()

    def exposure_provided_manually(self, exposure_estimate):
        # called by CaptureScript when user changes exposure using GUI
        # natively captured subs won't be estimated, so won't get updated
        for s in self.subs:
            if s.exposure_is_an_estimate:
                s.exposure = exposure_estimate
                s.exposure_is_manual_estimate = True
        # check if exposure has changed but only when we are not capturing natively
        if not self.is_empty() and self.subs[0].exposure_is_an_estimate:
            self.check_for_changes()

    def check_for_changes(self):
        # for light subs only, denote a change if any change in select/reject
        # for calibration frames self.changed is always True (so we can save master)
        if self.subs[0].sub_type == 'light':
            self.changed = {s.name for s in self.subs if s.status == 'reject'} != self.orig_rejects or \
                self.subs[0].exposure != self.original_exposure

    def realign(self, *args):
        self.recompute(realign=True)

    def recompute(self, widgy=None, realign=False):
        # initial load, recompute or realign
        Component.get('Aligner').reset()
        self.stack_cache = {}
        # shuffle-baased realign
        if realign:
            np.random.shuffle(self.subs)
            # if we have a B or Ha, keep shuffling until we start with that
            if len([s for s in self.subs if s.filter in {'B', 'Ha'}]) > 0:
                while self.subs[0].filter not in {'B', 'Ha'}:
                    np.random.shuffle(self.subs)
        self.selected_sub = -1
        Clock.schedule_once(self._reprocess_sub, 0)

    def _reprocess_sub(self, dt):
        if self.selected_sub + 1 < len(self.subs):
            self.process(self.subs[self.selected_sub + 1])
            self.selected_sub += 1
            Clock.schedule_once(self._reprocess_sub, 0)
        else:
            self.stack_changed()

    def process(self, sub):
        # Process sub on recompute/realign

        sub.image = None  # force reload
        Component.get('BadPixelMap').process_bpm(sub)
        if sub.sub_type == 'light':
            Component.get('Calibrator').calibrate(sub)
            Component.get('Aligner').process(sub)
        elif sub.sub_type == 'flat':
            Component.get('Calibrator').calibrate_flat(sub)




''' Virtual and real camera classes, all subclasses of GenericCamera.
    For the moment there is quite a lot of factorisable repetition
    but once development has slowed down that can be dealt with ;-)
'''

import time
import array
import os
from concurrent.futures import ThreadPoolExecutor
import numpy as np
from functools import partial

import zwoasi as asi

from kivy.logger import Logger
from kivy.app import App
from kivy.properties import BooleanProperty, ConfigParserProperty

from kivy.clock import Clock
from kivy.event import EventDispatcher

from jocular.component import Component


class GenericCamera(EventDispatcher):

    camera_connected = BooleanProperty(False)

    def __init__(self, controller, **kwargs):
        super().__init__(**kwargs)
        self.controller = controller

    def reset(self):
        pass
    def on_close(self):
        pass
    def pause(self):
        pass
    def get_image(self):
        return None
    def connect(self):
        pass
    def is_connected(self):
        return self.camera_connected
    def select(self):
        pass
    def deselect(self):
        pass
    def on_new_object(self):
        pass
    def capture_sub(self, **kwargs):
        return None


class ASICamera(GenericCamera):

    def reset(self):
        self.connect()
        self.last_capture = None
        self.pause()

    def on_new_object(self):
        if not Component.get('ObjectIO').existing_object:
            Component.get('Capture').enable_captures()

    def select(self):
        if not Component.get('ObjectIO').existing_object:
            self.on_new_object()

    def connect(self):
        if self.camera_connected:
            return
            
        self.camera_connected = False
        self._camera = None

        try:
            asipath = App.get_running_app().get_path('ASICamera2.dll')
            if not os.path.exists(asipath):
                Logger.error('Camera: cannot find ASICamera2.dll in resources')
                return
        except Exception as e:
            Logger.error('Camera: problem getting asipath ({:})'.format(e))
            return

        try:
            asi.init(os.path.abspath(asipath))
        except Exception as e:
            Logger.error('Camera: problem initialising asi ({:})'.format(e))
            return

        try:
            num_cameras = asi.get_num_cameras()
            if num_cameras == 0:
                Logger.error('Camera: no cameras found')
                return
        except Exception as e:
            Logger.error('Camera: problem getting number of cameras ({:})'.format(e))
            return

        # get first camera found
        try:
            self._camera = asi.Camera(0)
        except Exception as e:
            Logger.error('Camera: problem assigning camera 0 ({:})'.format(e))
            return

        # if we want camera info we can get it using
        # camera_info = self._camera.get_camera_property()

        # set some properties: in future will be options
        try:
            self._camera.set_control_value(asi.ASI_BANDWIDTHOVERLOAD, 
                self._camera.get_controls()['BandWidth']['MinValue'])
            self._camera.disable_dark_subtract()
            self._camera.set_image_type(asi.ASI_IMG_RAW16)
            self._camera.set_control_value(asi.ASI_GAIN, 150)
        except Exception as e:
            Logger.error('Camera: problem setting control values ({:})'.format(e))
            return

        Logger.info('Camera: connected!')
        self.camera_connected = True

    def handle_failure(self, message='camera problem'):
        Logger.error('Camera: handling failure {:}'.format(message))
        self.camera_connected = False  # to ensure reconnect if that was the issue
        self.last_capture = None
        if self.on_failure is not None:
            self.on_failure()

    def get_image(self):
        return self.last_capture

    def capture_sub(self, exposure=None, on_capture=None, on_failure=None,
        binning=None, internal_timing=False, return_image=False, is_bias=False):

        if is_bias:
            self.exposure = .001

        self.exposure = exposure
        self.on_failure = on_failure
        self.on_capture = on_capture

        # exposure is in microseconds
        self._camera.set_control_value(asi.ASI_EXPOSURE, int(self.exposure * 1e6))
        Logger.debug('Camera: starting exposure')
        self._camera.start_exposure()

        if return_image:
            time.sleep(self.exposure)
            ready = False
            while not ready:
                time.sleep(.2)
                ready = self._camera.get_exposure_status() != asi.ASI_EXP_WORKING
            if self._camera.get_exposure_status() != asi.ASI_EXP_SUCCESS:
                self.handle_failure('capture problem')
                return
            return self.get_camera_data()
             
        self.capture_event = Clock.schedule_once(self.check_exposure, 
            max(0.2, self.exposure))

    def get_camera_data(self):
        # from https://github.com/stevemarple/python-zwoasi/blob/master/zwoasi/__init__.py
        data = self._camera.get_data_after_exposure(None)
        whbi = self._camera.get_roi_format()
        shape = [whbi[1], whbi[0]]
        img = np.frombuffer(data, dtype=np.uint16)
        return img.reshape(shape) / 2 ** 16

    def check_exposure(self, *arg):
        if self._camera.get_exposure_status() != asi.ASI_EXP_WORKING:
            self.last_capture = self.get_camera_data()
            if self.on_capture is not None:
                Clock.schedule_once(self.on_capture, 0)
        else:
            # check exposure again in 200 ms
            self.capture_event = Clock.schedule_once(self.check_exposure, 0.2)

    def pause(self):
        # Cancel any pending reads
        if hasattr(self, '_camera') and self._camera is not None:
            self._camera.stop_exposure()
        if hasattr(self, 'capture_event'):
            self.capture_event.cancel()



class AscomCamera(GenericCamera):

    def reset(self):
        self.connect()
        self.last_capture = None
        self.pause()

    def on_new_object(self):
        if not Component.get('ObjectIO').existing_object:
            Component.get('Capture').enable_captures()

    def select(self):
        if not Component.get('ObjectIO').existing_object:
            self.on_new_object()

    def connect(self):
        if self.camera_connected:
            if hasattr(self, '_camera') and self._camera.connected:
                return
            
        self.camera_connected = False
        self._camera = None

        try:
            import win32com.client
        except Exception as e:
            Logger.error('Camera: cannot import win30com.client: ({:})'.format(e))
            return

        try:
            x = win32com.client.Dispatch("ASCOM.Utilities.Chooser")
            x.DeviceType = 'Camera'
            driver = x.Choose(None)
            Logger.info('Camera: got driver {:}'.format(driver))
            camera = win32com.client.Dispatch(driver)
            Logger.info('Camera: dispatched driver {:}'.format(camera))
            self._camera = camera
            Logger.info('Camera: assigned driver to _camera')
            self._camera.connected = True   # this is camera-specific prop
            Logger.info('Camera: set connected on driver to True')
            self.camera_connected = True    # note this is generic prop
        except Exception as e:
            Logger.error('Camera: problem connecting ASCOM camera: ({:})'.format(e))
            return

        Logger.info('Camera: ASCOM camera connected')

    def handle_failure(self, message='camera problem'):
        Logger.error('Camera: handling failure {:}'.format(message))
        self.camera_connected = False  # to ensure reconnect if that was the issue
        self.last_capture = None
        if self.on_failure is not None:
            self.on_failure()

    def get_image(self):
        return self.last_capture

    def capture_sub(self, exposure=None, on_capture=None, on_failure=None,
        binning=None, internal_timing=False, return_image=False, is_bias=False):

        self.exposure = exposure
        self.on_failure = on_failure
        self.on_capture = on_capture

        if is_bias:
            self.exposure = .001
            openshutter = False
        else:
            openshutter = True

        self._camera.startExposure(self.exposure, openshutter) 

        if return_image:
            time.sleep(self.exposure)
            ready = False
            while not ready:
                time.sleep(.2)
                ready = self._camera.ImageReady
            return self.get_camera_data()

        self.capture_event = Clock.schedule_once(self.check_exposure, 
            max(0.2, self.exposure))

    def check_exposure(self, *arg):
        if self._camera.ImageReady:
            self.last_capture = self.get_camera_data()
            if self.on_capture is not None:
                Clock.schedule_once(self.on_capture, 0)
        else:
            # check exposure again in 200 ms
            self.capture_event = Clock.schedule_once(self.check_exposure, 0.2)

    def get_camera_data(self):
        # data is ready so get it, convert to correct size, and scale to range 0-1
        # assumes 16 bits for now
        im = np.array(self._camera.ImageArray)
        shape = [self._camera.CameraXSize, self._camera.CameraYSize]
        return im.reshape(shape) / 2 ** 16

    def pause(self):
        # Cancel any pending reads
        Logger.debug('Camera: pause called')
        if hasattr(self, '_camera') and self._camera is not None:
            # occasionally objects to being aborted
            if self._camera.CanStopExposure:
                Logger.debug('Camera: can stop exposure, trying')
                try:
                    self._camera.StopExposure()
                    Logger.debug('Camera: stopped exposure')
                except Exception as e:
                    Logger.error('Camera: unable to stop exposure ({:})'.format(e))
        if hasattr(self, 'capture_event'):
            self.capture_event.cancel()


''' ------------ SX cameras (just Lodestar for now)
'''

SX_CLEAR_PIXELS = 1
SX_READ_PIXELS_DELAYED = 2
SX_READ_PIXELS = 3
SX_RESET = 6

# convert int formats
def convert_int(x, from_type='uint16', to_type='uint8'):
    return np.array([x], dtype=from_type).view(dtype=to_type)

class SXCamera(GenericCamera):

    def reset(self):
        self.connect()
        self.last_capture = None
        self.pause()

    def on_new_object(self):
        if not Component.get('ObjectIO').existing_object:
            Component.get('Capture').enable_captures()

    def select(self):
        if not Component.get('ObjectIO').existing_object:
            self.on_new_object()

    def pause(self):
        # Cancel any pending reads
        if hasattr(self, 'capture_event'):
            if not self.internal_timing:  
                # don't cancel reading if using internal timing (causes crash on LS)
                self.capture_event.cancel()

    def clean_up(self):
        if hasattr(self, 'lodestar'):
            import usb.util
            usb.util.dispose_resources(self.lodestar)
            Logger.debug('Camera: disposed')

    def handle_failure(self, message='camera problem'):
        Logger.error('Camera: handle_failure {:}'.format(message))
        self.camera_connected = False  # to ensure reconnect if that was the issue
        self.last_capture = None
        try:
            self.clean_up()
        except Exception as e:
            Logger.warn('Camera: problem disposing {:}'.format(e))
            self.contoller.warn('disconnected?')
        finally:
            if self.on_failure is not None:
                self.on_failure()

    def get_image(self):
        return self.last_capture

    def capture_sub(self, exposure=None, on_capture=None, on_failure=None,
        binning=None, internal_timing=False, return_image=False, is_bias=False):

        if is_bias:
            self.capture_bias(on_capture=on_capture, on_failure=on_failure)
            return

        # store these so they can be used without argument passing in futures etc
        self.on_capture = on_capture
        self.on_failure = on_failure
        self.exposure = exposure
        self.binning = binning
        self.internal_timing = internal_timing
        self.return_image = return_image
        self.last_capture = None

        # try to connect to camera (will handle if already connected)
        self.connect()

        if not self.camera_connected:
            self.handle_failure('camera not connected')
            return

        try:
            if internal_timing:
                # do internal_timed exposure
                self.start_internal_exposure()
                if return_image:
                    time.sleep(exposure)
                    return self.lodestar_read()
                else:
                    self.capture_event = Clock.schedule_once(
                        self.read_internal_exposure, max(0.1, exposure)
                    )
            else:
                # externally-timed
                self.start_external_exposure()
                if return_image:
                    time.sleep(exposure)
                    return self.read_external_exposure()
                else:
                    self.capture_event = Clock.schedule_once(
                        self.read_external_exposure, max(0.2, self.exposure)
                    )

        except Exception as e:
            self.handle_failure('capture problem {:}'.format(e))

    def capture_bias(self, on_capture=None, on_failure=None):
        # special case for bias as it needs even and odd for internally-timed exposure
        self.on_capture = on_capture
        self.on_failure = on_failure
        try:
            self.connect()
            expo = 0.001
            self.exposure_command(rows='odd', exposure=expo)
            time.sleep(expo)
            odd_pixels = self.lodestar.read(0x82, self.height * self.width, 10000)
            self.exposure_command(rows='even', exposure=expo)
            time.sleep(expo)
            even_pixels = self.lodestar.read(0x82, self.height * self.width, 10000)
            pool = ThreadPoolExecutor(3)
            future = pool.submit(
                partial(self.deinterlace, odd_pixels, even_pixels)
            )  # thread handles read
            future.add_done_callback(
                self.image_ready
            )  # when future is done, call image_ready
        except Exception as e:
            self.handle_failure('problem in capture_bias {:}'.format(e))

    def exposure_command(self, rows='odd', exposure=None):
        # generate write command for sensor
        try:
            rowcode = {'odd': 1, 'even': 2, 'all': 3}
            cmd = SX_READ_PIXELS if exposure is None else SX_READ_PIXELS_DELAYED
            params = array.array(
                'B',
                [
                    0x40, cmd, rowcode[rows], 0, 0, 0, 10, 0, 0, 0, 0, 0,
                    self.lswidth, self.mswidth, self.lsheight, self.msheight, 1, 1
                ],
            )
            if exposure is not None:
                # add exposure details to command
                expo = int(exposure * 1000)  #  ms
                exp1, exp2, exp3, exp4 = convert_int(expo, from_type='uint32')
                params = params + array.array('B', [exp1, exp2, exp3, exp4])
            self.lodestar.write(1, params, 1000)
        except Exception as e:
            self.handle_failure('problem in exposure_command {:}'.format(e))

    def start_internal_exposure(self):
        try:
            self.exposure_command(rows='odd', exposure=self.exposure)
        except Exception as e:
            self.handle_failure('problem in start_internal_exposure {:}'.format(e))

    def start_external_exposure(self):
        try:
            self.lodestar.ctrl_transfer(0x40, SX_RESET, 0, 0, 0)
            self.lodestar.ctrl_transfer(0x40, SX_CLEAR_PIXELS, 0, 0, 0)
            Logger.trace('Camera: started external exposure')
        except Exception as e:
            self.handle_failure('problem in start_external_exposure {:}'.format(e))

    def read_internal_exposure(self, *args):
        try:
            pool = ThreadPoolExecutor(3)
            future = pool.submit(self.lodestar_read)  # thread handles read
            future.add_done_callback(self.image_ready)  # when future is done, call image_ready
        except Exception as e:
            self.handle_failure('problem in read_internal_exposure {:}'.format(e))

    def read_external_exposure(self, *args):
        # Reads sensor at end of exposure time

        try:
            self.exposure_command(rows='odd')
            odds = self.lodestar.read(0x82, self.height * self.width, timeout=10000)
            self.exposure_command(rows='even')
            evens = self.lodestar.read(0x82, self.height * self.width, timeout=10000)

            # de-interlace to create final image
            self.last_capture = self.deinterlace(odds, evens)

            if self.on_capture is not None:
                self.on_capture()
        except Exception as e:
            self.handle_failure('problem in read_external_exposure {:}'.format(e))

    def lodestar_read(self):
        try:
            odd_pixels = self.lodestar.read(0x82, self.height * self.width, 10000)
            return self.deinterlace(odd_pixels, odd_pixels)
        except Exception as e:
            self.handle_failure('problem in lodestar_read {:}'.format(e))

    def image_ready(self, future):
        self.last_capture = future.result()  # store image read from sensor
        if self.on_capture is not None:
            Clock.schedule_once(self.on_capture, 0)

    def deinterlace(self, odd8, even8):
        '''De-interlaces Lodestar.'''

        # convert uint8 arrays to uint16 and reshape
        odd = np.frombuffer(odd8, dtype='uint16').reshape(self.half_height, self.width)
        even = np.frombuffer(even8, dtype='uint16').reshape(self.half_height, self.width)

        # generate new array (full height)
        pix = np.zeros((self.height, self.width))

        # insert odd rows
        pix[::2, :] = odd

        # insert even rows and normalise to account for slight delay in reading
        pix[1::2, :] = even * (np.mean(odd) / np.mean(even))

        return pix / 2 ** 16

    def connect(self):
        ''' Connect to Lodestar. For other SX cameras will need to change idProduct
            and check interface/endpoint details. Not working on Windows.
        '''

        if self.camera_connected:
            return

        if os.name == 'nt':
            self.info('not supported')
            return

        self.camera_connected = False

        try:
            import usb.core
            import usb.util
        except:
            self.controller.warn('no usb.core or usb.util')
            return

        self.lodestar = None
        try:
            self.lodestar = usb.core.find(idVendor=0x1278, idProduct=0x0507)
        except:
            self.controller.warn('cannot find camera')

        if self.lodestar is None:
            self.controller.info('not connected')
            return

        try:
            conf = self.lodestar.get_active_configuration()
        except Exception as e:
            Logger.warn('Camera: cannot get config ({:})'.format(e))
            self.controller.warn('no config')
            self.clean_up()
            return

        self.camera_connected = True
        self.controller.info('SX Lodestar')

        # used in capture; note h // 2 for interlaced; needs changing for ultrastar no doubt
        self.width, self.height = 752, 580
        self.half_height = self.height // 2
        self.lswidth, self.mswidth = convert_int(self.width)
        self.lsheight, self.msheight = convert_int(self.half_height)




''' This is a watched camera that is always on i.e. like the
    original watched folder arrangement. Have to ensure we disable
    the capture button and set capturing in Watcher to True throughout.
'''

class WatchedCamera(GenericCamera):
 
    def select(self):
        self.camera_connected = True    # effectively always true
        self.on_new_object()

    def deselect(self):
        ''' Stop watcher and capture capturing; any new camera mode selected
            is responsible for enabling capture controls
        '''
        Component.get('Watcher').capturing = False
        Component.get('Capture').capturing = False

    def on_new_object(self):
        # allow watcher to capture only if not viewing previous object
        Component.get('Capture').disable_captures()
        Component.get('Watcher').capturing = not Component.get('ObjectIO').existing_object   


''' A watched camera that responds to the capture start/stop toggle. Subs
    are only allowed through when capturing. Behaviour controlled by Watcher.
'''

class WatchedCameraOnOff(GenericCamera):

    def connect(self):
        self.on_new_object()

    def capture_sub(self, exposure=None, on_capture=None, on_failure=None, binning=None, 
                    internal_timing=False, return_image=False, is_bias=False):
        Component.get('Watcher').capturing = True

    def pause(self):
        Component.get('Watcher').capturing = False

    def reset(self):
        self.on_new_object()

    def select(self):
        self.on_new_object()

    def deselect(self):
        Component.get('Capture').disable_captures()

    def on_new_object(self):
        # allow watcher to capture only if not viewing previous object
        self.camera_connected = not Component.get('ObjectIO').existing_object
        if self.camera_connected:
            Component.get('Capture').enable_captures()
        




''' Convenient place to support ring metrics.
'''

from kivy.metrics import dp
from kivy.logger import Logger
from kivy.core.window import Window

class Metrics():

    metrics = {}
    width = 0
    height = 0

    @classmethod
    def get(cls, p):
        width, height = Window.size
        if (cls.width != width) or (cls.height != height):
            cls.update()
        if p in cls.metrics:
            return cls.metrics[p]
        else:
            Logger.error('Metrics: no such metric {:}'.format(p))

    @classmethod
    def update(cls):

        width, height = Window.size

        origin = .5 * width, .5 * height
        radius = .5 * height if width > height else .5 * width
        diag = (origin[0] ** 2 +origin[1] ** 2) ** .5

        # scope rings: 3 + outer (obscuring) ring
        inner_width = dp(8)
        middle_width = dp(28)
        outer_width = dp(22)

        radii = {}
        radii['background'] = radius, diag
        radii['outer'] = radii['background'][0] - outer_width, radii['background'][0]
        radii['middle'] = radii['outer'][0] - middle_width, radii['outer'][0]
        radii['inner'] = radii['middle'][0] - inner_width, radii['middle'][0]
        radii['image'] = 0, radii['inner'][0]

        ring_radius = {nm: (r_inner + r_outer) / 2 for nm, (r_inner, r_outer) in radii.items()}

        ring_thickness = {nm: (r_outer - r_inner) / 2 for nm, (r_inner, r_outer) in radii.items()} 
        inner_ring = radii['middle'][0] + middle_width / 2
        outer_ring = radii['outer'][0] + outer_width / 2
        outside_ring = radii['outer'][1] + outer_width / 2 

        ring_radius['capture_ring'] = (inner_ring + outer_ring - inner_width)/ 2
        ring_thickness['capture_ring'] = (outer_ring - inner_ring)

        cls.width = width
        cls.height = height

        cls.metrics = {
            'origin': origin,
            'radius': radius,
            'inner_radius': radii['inner'][0],
            'inner_ring': inner_ring, 
            'outer_ring': outer_ring, 
            'outside_ring': outside_ring,
            'fine_control': outside_ring + dp(20),
            'mid_ring': (inner_ring + outer_ring)/ 2,
            'outer_width': outer_width,
            'radii': radii,
            'ring_radius': ring_radius,
            'ring_thickness': ring_thickness
        }



''' Monitors any incoming subs
'''

import os
import numpy as np

from kivy.logger import Logger
from kivy.app import App
from kivy.clock import Clock
from kivy.properties import ConfigParserProperty, BooleanProperty

from skimage.transform import rescale, downscale_local_mean

from jocular.component import Component
from jocular.utils import move_to_dir
from jocular.image import Image, ImageNotReadyException, is_fit, save_image

class Watcher(Component):

    # new in v0.5
    bayerpattern = ConfigParserProperty('mono', 'Watcher', 'bayerpattern', 'app', val_type=str)
    binfac_on_input = ConfigParserProperty(1, 'Watcher', 'binfac_on_input', 'app', val_type=int)
    binmethod = ConfigParserProperty('interpolate', 'Watcher', 'binmethod', 'app', val_type=str)
    save_originals = ConfigParserProperty(1, 'Watcher', 'save_originals', 'app', val_type=int)
    
    # set to True if capture mode is Watched and user is capturing
    capturing = BooleanProperty(False)

    def __init__(self):
        super(Watcher, self).__init__()
        self.watched_dir = App.get_running_app().get_path('watched')
        self.watching_event = Clock.schedule_interval(self.watch, .3)

    def on_new_object(self):
        self.flush()

    def on_close(self):
        self.watching_event.cancel()
        self.flush()    # user has had enough so move any FITs to delete

    def flush(self):
        # move any FITs that are not masters to a 'unused' folder in watcher
        # (masters are saved immediately before new object, so risk they are not spotted in time!)
        for f in os.listdir(self.watched_dir):
            if is_fit(f):
                if not f.startswith('master'):
                    move_to_dir(os.path.join(self.watched_dir, f), 'unused')

    def get_possible_fits(self):
        fits = [os.path.join(self.watched_dir, d) for d in os.listdir(self.watched_dir)]
        # for ASIlive look deeper
        asipath = os.path.join(self.watched_dir, 'ASILive_AutoSave', 'SingleFrame')
        if os.path.exists(asipath):
            for sdir in os.listdir(asipath):
                pth = os.path.join(asipath, sdir)
                if os.path.isdir(pth):
                    fits += [os.path.join(pth, d) for d in os.listdir(pth)]
        return fits

    def watch(self, dt):
        '''Watcher handles two types of event:

            1.  Users drop individual subs (manually or via a capture program); if
                validated these are passed to ObjectIO
            2.  Users drop master calibration frames; if validated, passed to Calibrator
                to become part of the library and become available for immediate use

            Validated files are removed from 'watched'. If not validated, files are
            either left to be re-handled on the next event cycle, or moved to 'invalid'.
            Non-fits files are moved to 'ignored'
        '''

        ''' New approach with WatchedCamera is to allow through all Jocular-created subs
            but only allow through alien subs if WatchedCamera is capturing. Rest are
            shunted off to ignored.
        '''

        for path in self.get_possible_fits():
            f = os.path.basename(path)
            if is_fit(f):
                try:
                    s = Image(path, check_image_data=True)

                    # captured natively so let it through
                    if s.created_by_jocular:
                        if s.is_master:
                            Component.get('Calibrator').new_master_from_watcher(s)
                        else:
                            Component.get('ObjectIO').new_sub_from_watcher(s)

                    # watcher is not allowing stuff in
                    elif not self.capturing:
                        move_to_dir(path, 'ignored')

                    # non-jocular FITs are converted to Jocular FITs, spat back out to Watched
                    # directory, and the originals saved if required
                    else:
                        self.process_alien_sub(s, path)
 
                except ImageNotReadyException as e:
                    # give it another chance on next event cycle
                    Logger.debug('Watcher: image not ready {:} ({:})'.format(f, e))

                # added in v0.4.5 to catch error thrown by new_master    
                except IOError as e:
                    Logger.debug('Watcher: {:}'.format(e))
                    move_to_dir(path, 'invalid')

                except Exception as e:
                    # irrecoverable, so move to invalid, adding timestamp
                    Logger.debug('Watcher: other issue {:} ({:})'.format(f, e))
                    move_to_dir(path, 'invalid')

            elif not os.path.isdir(path):
                move_to_dir(path, 'nonfits')


    def bin(self, im):
        binfac = int(self.binfac_on_input)
        if binfac < 2:
            return im
        if self.binmethod == 'interpolation':
            return rescale(im, 1 / binfac, anti_aliasing=True, mode='constant', 
                preserve_range=True, multichannel=False)
        return downscale_local_mean(im, (binfac, binfac))

    def process_alien_sub(self, s, path):

        bn = os.path.basename(path)
        im = s.get_image()

        # don't debayer if mono or if it is a master calibration frame
        if self.bayerpattern == 'mono' or s.is_master:
            self.save_mono(s, self.bin(im), bn, filt=None)

        # debayer
        else:
            from colour_demosaicing import demosaicing_CFA_Bayer_bilinear
            rgb = demosaicing_CFA_Bayer_bilinear(im, pattern=self.bayerpattern)
            # rescale to original intensity range
            cfa_min, cfa_max = np.min(im), np.max(im)
            rgb_min, rgb_max = np.min(rgb), np.max(rgb)
            rgb = cfa_min + (cfa_max - cfa_min) * (rgb - rgb_min) / (rgb_max - rgb_min)
            for i, chan in enumerate(['R', 'G', 'B']):
                self.save_mono(s, self.bin(rgb[:, :, i]), bn, filt=chan)
 
        if self.save_originals and path is not None:
            Component.get('ObjectIO').save_original(path)


    def save_mono(self, sub, im, nm, filt=None):
        ''' Save image, binning if required, constructing details from Image instance
        '''

        details = Component.get('Capture').get_capture_details()

        filt = sub.filter if filt is None else filt
        filt = details['filter'] if filt is None else filt
        sub_type = details['sub_type'] #  if sub.sub_type is None else sub.sub_type
        exposure = details['exposure'] if sub.exposure is None else sub.exposure 
        temperature = sub.temperature

        if sub.is_master:
            save_image(data=im,
                path=os.path.join(self.watched_dir, 'master{:}.fit'.format(sub_type)),
                exposure=exposure, filt=filt, temperature=temperature, sub_type='master ' + sub_type,
                nsubs=sub.nsubs)
        else:
            save_image(data=im.astype(np.uint16), 
                path=os.path.join(self.watched_dir, '{:}_{:}'.format(filt, nm)),
                filt=filt, sub_type=sub_type, exposure=exposure, temperature=temperature)



''' Initiates jocular, checking if we have a data directory
'''

import os
import sys
import click
import glob
import json
from datetime import datetime
from pathlib import Path

from jocular import __version__

def get_datadir():
    try:
        with open(os.path.join(str(Path.home()), '.jocular'), 'r') as f:
            return f.read().strip()
    except:
        return None

def print_version(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return
    click.echo('Version {:}'.format(__version__))
    ctx.exit()

@click.command()
@click.option(
    '--datadir',
    type=click.Path(exists=True),
    default=get_datadir(),
    show_default=True,
    help='where to store Jocular data',
)
@click.option(
    '--log',
    show_default=True,
    type=click.Choice(['error', 'warning', 'debug'], case_sensitive=False),
    default='error',
)
@click.option('--version', '-v', is_flag=True, callback=print_version,
              expose_value=False, is_eager=True)
@click.option('--rebuildobservations', '-r', is_flag=True, default=False,
    show_default=True,
    expose_value=True, is_eager=True,
    help='Construct observations file by scanning captures')
def startjocular(datadir, log, rebuildobservations):

    # stop Kivy from interpreting any args
    os.environ["KIVY_NO_ARGS"] = "1"

    if datadir is None:
        print()
        print('  Jocular cannot find your data directory. This might be because')
        print('  this is the first time you have used Jocular (welcome!), or')
        print('  because you have moved it. Please specify it using the --datadir')
        print('  option e.g.')
        print('  ')
        print('    jocular --datadir /Users/Martin/JocularData')
        print('  ')
        print('  Note that the directory must already exist, so if you are')
        print('  starting to use Jocular for the first time, please create')
        print('  the directory before running the command. You will not need')
        print('  to specify the datadir on subsequent runs of Jocular unless')
        print('  you have changed its location, in which case you will see')
        print('  this message.')
        print()
        sys.exit()

    # ensure we are using absolute paths so that user can start from any place
    datadir = os.path.abspath(datadir)

    # store absolute path to datadir in .jocular
    try:
        with open(os.path.join(str(Path.home()), '.jocular'), 'w') as f:
            f.write(datadir)
    except Exception as e:
        sys.exit(
            'Unable to write .jocular in home directory {:} ({:})'.format(
                str(Path.home()), e
            )
        )

    # timestamp in datadir, which checks it is writeable
    try:
        with open(os.path.join(datadir, '.lastwritten'), 'w') as f:
            f.write(datetime.now().strftime('%Y-%m-%d %H:%M:%S'))
    except Exception as e:
        sys.exit('Unable to write to user data directory {:} ({:})'.format(datadir, e))

    # if user wishes to rebuild observations, or we cannot find .json
    obspath = os.path.join(datadir, 'previous_observations.json')
    if rebuildobservations or not os.path.exists(obspath):
        try:
            from jocular.utils import move_to_dir
            from jocular.metadata import get_metadata

            # move prev to deleted in case user wishes to restore
            pth = os.path.join(datadir, 'previous_observations.json')
            if os.path.exists(obspath):
                move_to_dir(obspath, os.path.join(datadir, 'deleted'))

            # extract all observational data from captures
            observations = {}
            for sesh in glob.glob(os.path.join(datadir, 'captures', '*')):
                for dso in glob.glob(os.path.join(sesh, '*')):
                    if os.path.isdir(dso):
                        observations[dso] = get_metadata(dso, simple=True)

            # save
            with open(obspath, 'w') as f:
                json.dump(observations, f, indent=1)


        except Exception as e:
            print('Problem rebuilding previous observations ({:})'.format(e))
            sys.exit()

    # write critical kivy preferences
    from kivy.config import Config
    from jocular.jocular import Jocular

    Config.set('kivy', 'log_level', log)
    Config.set('kivy', 'keyboard_mode', 'system')
    Config.set('kivy', 'log_enable', 0)
    Config.set('kivy', 'exit_on_escape', '0')
    Config.set('graphics', 'position', 'auto')
    Config.set('graphics', 'fullscreen', 0)
    Config.set('graphics', 'borderless', 0)
    Config.set('postproc', 'double_tap_time', 250)
    Config.set('input', 'mouse', 'mouse,multitouch_on_demand')
    Config.write()

    # finally we can start app
    try:
        Jocular().run()
    except Exception as e:
        print('Jocular failed with error {:}'.format(e))
        sys.exit()



''' Handles the GUI elements of getting info about different capture scripts.
    Capturing itself is done with the Capture component
'''

import json
import math
from functools import partial

from kivy.app import App
from kivy.properties import OptionProperty, ListProperty, NumericProperty, ObjectProperty, StringProperty, BooleanProperty
from kivy.lang import Builder
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.gridlayout import GridLayout
from kivy.uix.floatlayout import FloatLayout
from kivy.uix.togglebutton import ToggleButton
from kivy.uix.label import Label
from kivy.uix.textinput import TextInput
from kivy.metrics import dp

from jocular.component import Component
from jocular.widgets import JPopup

Builder.load_string('''

<MyLabel>:
    size_hint: 1, None
    height: dp(24)
    text_size: self.size
    markup: True
    halign: 'left'

<ExposureToggle>:
    markup: True
    background_color: 0, 0, 0, 0
    color: app.highlight_color if self.state == 'down' else app.lowlight_color
    font_size: '17sp'
    group: 'expos'

<FilterToggle>:
    canvas.before:
        Color:
            rgba: .8, .8, .8, .5
        Ellipse:
            pos: self.pos
            size: self.width, self.height
        Color:
            rgb: self.filter_color
            a: 1 if self.state == 'down' else .3
        Ellipse:
            pos: self.x + 2, self.y + 2
            size: self.width - 5, self.height - 5
    markup: True
    color: (1, 1, 1, 1) if self.state == 'down' else (0, 0, 0, 1)
    background_color: 0, 0, 0, 0
    size: dp(50), dp(50)
    font_size: '18sp' if self.state == 'down' else '16sp'
    size_hint: None, None

<ChooseMultiple>:
    orientation: 'vertical'
    grid: _grid
    size_hint: None, None
    width: dp(240)
    spacing: dp(0), dp(15)
    height: _grid.height + _subs_box.height + _temp_box.height + _confirm_box.height + dp(10)

    BoxLayout:
        id: _grid
        size_hint: 1, None
        height: dp(280)

    BoxLayout:
        id: _subs_box
        size_hint: 1, None
        orientation: 'vertical'
        height: dp(80) if root.show_subs else dp(0)
        opacity: 1 if root.show_subs else 0
        index: 1 if root.show_subs else 100
        disabled: not root.show_subs

        Label:
            color: .5, .5, .5, 1
            font_size: '20sp'
            text: '{:} subs per filter'.format(_nsubs.value)

        Slider:
            size_hint: 1, None
            height: dp(30)
            id: _nsubs
            min: 2
            max: root.max_subs
            value: root.capture.scripts[root.capture.current_script].get('nsubs', 0)
            step: 1

    BoxLayout:
        id: _temp_box
        size_hint: 1, None
        height: dp(80) if root.show_temperature else dp(0)
        opacity: 1 if root.show_temperature else 0
        disabled: not root.show_temperature
        orientation: 'vertical'

        Label:
            text: 'Temperature {:}\N{DEGREE SIGN}C'.format(_temperature.value) if _temperature.value > -26 else 'Temperature not set'
            color: .5, .5, .5, 1
            font_size: '20sp'
        Slider:
            size_hint: 1, None
            height: dp(30)
            id: _temperature
            value: root.session.temperature if root.session.temperature is not None else -26
            on_value: root.session.temperature = self.value if self.value > -26 else None
            min: -26
            max: 40
            step: 1
            
    BoxLayout:
        id: _confirm_box
        size_hint: 1, None
        height: dp(40) # if root.show_confirm else dp(0)
        opacity: 1     # if root.show_confirm else 0

        Button:
            text: 'Done'
            size_hint: .5, .8
            on_press: root.capture.exposure_done() if root.choice_type == 'exposure' else root.capture.filters_done(_nsubs.value)
        Button:
            text: 'Cancel'
            size_hint: .5, .8
            on_press: root.capture.cancel_multiple(root.choice_type)
''')

class FilterToggle(ToggleButton):
    filter_color = ListProperty([1, 1, 1, 1])

class ChooseMultiple(BoxLayout):
    grid = ObjectProperty()
    choice_type = StringProperty('exposure')
    show_temperature = BooleanProperty(False)
    show_subs = BooleanProperty(False)
    subs = NumericProperty(2)
    max_subs = NumericProperty(2)

    def __init__(self, capture=None, session=None, choice_type=None, **kwargs):
        script = capture.current_script
        self.show_temperature = script == 'dark'
        self.max_subs = 20
        self.show_subs = script == 'seq'
        self.capture = capture
        self.session = session
        self.choice_type = choice_type
        super().__init__(**kwargs)

class ExposureToggle(ToggleButton):
    pass

def exp_to_str(e):
    if e < .1:
        return '{:.0f} ms'.format(e*1000)
    if e < 1:
        return '{:.2f}s'.format(e)
    if e < 10:
        if abs(int(e) - e) < .0001:
            return '{:.0f}s'.format(e)
        return '{:.2f}s'.format(e)
    return '{:.0f}s'.format(e)

def str_to_exp(s):
    s = s.strip()
    try:
        if s.endswith('ms'):
            return float(s[:-2].strip()) / 1000
        if s.endswith('s'):
            return float(s[:-1].strip())
        return float(s)
    except:
        raise Exception


class CaptureScript(Component):

    current_script = OptionProperty('frame', options=['light', 'frame', 'dark', 'flat', 'seq', 'bias'])    

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.app = App.get_running_app()
        try:
            with open(self.app.get_path('capture_scripts.json'), 'r') as f:
                self.scripts = json.load(f)
        except:
            # no capture scripts in user directory so use shipped version to start with
            with open(self.app.get_path('shipped_capture_scripts.json'), 'r') as f:
                self.scripts = json.load(f)

        self.on_current_script()

    def on_close(self, *args):
        with open(self.app.get_path('capture_scripts.json'), 'w') as f:
            json.dump(self.scripts, f, indent=1)       

    def on_current_script(self, *args):
        script = self.scripts[self.current_script]
        self.set_exposure_button(script['exposure'])
        self.app.gui.set('filter_button', script['filter'][0])
        self.reset_generator()
        self.app.gui.set('show_reticle', False, update_property=True)
        self.app.gui.set('mean', True, update_property=True)
        if self.current_script == 'flat':
            self.app.gui.set('80', True, update_property=True)
        elif self.current_script == 'frame':
            self.app.gui.set('show_reticle', True, update_property=True)

    def set_exposure_button(self, val):
        self.app.gui.set('exposure_button', exp_to_str(val))

    def get_current_exposure(self, sub_type):
        if sub_type in self.scripts:
            return self.scripts[sub_type]['exposure']
        return None

    def show_exposure_panel(self):
        ''' New in v0.4.6: allow user to type exposure as well as presets
        '''

        if self.current_script in {'flat', 'bias'}:
            return

        content = BoxLayout(size_hint=(None, None), size=(dp(240), dp(300)), 
            orientation='vertical',
            spacing=(dp(0), dp(20))
            )

        grid = GridLayout(size_hint=(None, None), size=(dp(240), dp(240)), cols=4,
            spacing=(dp(5), dp(0)))

        tinput = TextInput(
            text=exp_to_str(self.scripts[self.current_script].get('exposure', 0)),
            size_hint=(1, 1),
            height=dp(24),
            multiline=False,
            font_size='19sp',
            hint_text='type exposure or select preset',
            background_color=[.8, .8, .8, 1])
        tinput.valign = 'top'
        tinput.bind(on_text_validate=self.exposure_selected_from_text)
        content.add_widget(tinput)
        content.add_widget(Label(size_hint=(None, None), size=(dp(240), dp(20))))
        content.add_widget(grid)

        script = self.scripts[self.current_script]

        for e in script['expos']:
            grid.add_widget(ExposureToggle(
                text=exp_to_str(e), 
                state='down' if e == script['exposure'] else 'normal',
                on_press=partial(self.exposure_selected, e, tinput)))

        if self.current_script == 'dark':
            panel_content = ChooseMultiple(choice_type='exposure', capture=self, session=Component.get('Session'))
            panel_content.grid.add_widget(content)
            self.popup = JPopup(title='Choose exposure and temperature', content=panel_content)
        else:
            self.popup = JPopup(title='Choose exposure', content=content)

        self.popup.open()


    def exposure_selected_from_text(self, textinput):
        try:
            expo = str_to_exp(textinput.text)
            self.app.gui.set('exposure_button', exp_to_str(expo))
            self.scripts[self.current_script]['exposure'] = expo
            # don't dismiss popup if dark as user has opportunity to select temperature too
            if self.current_script != 'dark':
                self.exposure_done()
        except Exception as e:
            textinput.foreground_color[0] = 1

    def set_external_details(self, exposure=None, sub_type=None, filt=None):
        self.current_script = sub_type
        script = self.scripts[self.current_script]
        if exposure is not None:
            script['exposure'] = exposure
            self.app.gui.set('exposure_button', exp_to_str(exposure))
        if filt is not None:
            self.app.gui.set('filter_button', filt)

    def exposure_selected(self, expo, tinput, expo_toggle=None):
        self.app.gui.set('exposure_button', exp_to_str(expo))
        self.scripts[self.current_script]['exposure'] = expo
        tinput.text = exp_to_str(expo)
        # don't dismiss popup if dark as user has opportunity to select temperature too
        if self.current_script != 'dark':
            self.exposure_done()

    def exposure_done(self, *args):
        # user has provided exposure
        self.reset_generator()
        self.popup.dismiss()
        # tell stack about this setting so that it can propagate to subs where required
        Component.get('Stacker').exposure_provided_manually(self.scripts[self.current_script]['exposure'])

    def filters_done(self, nsubs=None, *args):
        script = self.scripts[self.current_script]
        script['nsubs'] = nsubs
        self.reset_generator()
        self.popup.dismiss()

    def cancel_multiple(self, typ=None, *args):
        self.popup.dismiss()

    def show_filter_panel(self):

        fw = Component.get('FilterWheel')
        non_empties = {f for f in fw.filters if f != 'empty'}

        if len(non_empties) < 2:
            return
        if self.current_script in {'dark', 'bias'}:
            return

        script = self.scripts[self.current_script]

        fw_layout = FloatLayout(size_hint=(None, None), size=(dp(240), dp(240)))
        rads = math.pi / 180
        for ang, f in zip([0, 40, 80, 120, 160, 200, 240, 280, 320], fw.filters):
            fw_layout.add_widget(FilterToggle(text=f, filter_color=fw.get_color(f),
                on_press=partial(self.filter_selected, f),
                state='down' if f in script['filter'] else 'normal',
                pos_hint={
                    'center_x': ( .8 * math.cos(ang * rads) + 1 ) / 2, 
                    'center_y': ( .8 * math.sin(ang * rads) + 1 ) / 2}))

        if self.current_script != 'seq':
            self.popup = JPopup(title='Choose filter', content=fw_layout)
        else:
            content = ChooseMultiple(choice_type='filter', capture=self, session=Component.get('Session'))
            content.grid.add_widget(fw_layout)
            self.popup = JPopup(title='Choose multiple filters', content=content)
        self.popup.open()

    def filter_selected(self, f, filter_toggle):
        if f == '-':
            return
        script = self.scripts[self.current_script]
        if self.current_script != 'seq':
            self.app.gui.set('filter_button', f)
            script['filter'] = [f]
            self.reset_generator()
            self.popup.dismiss()
        else:
            if filter_toggle.state == 'down':
                script['filter'].append(f)
            else:
                script['filter'].remove(f)

    def light_generator(self):
        script = self.scripts['light']
        yield 'set filter', script['filter'][0]
        yield 'set exposure', script['exposure']
        while True:
            yield 'expose long'

    def seq_generator(self):
        script = self.scripts['seq']
        yield 'set exposure', script['exposure']
        while True:
            for f in Component.get('FilterWheel').order_by_transmission(script['filter']):
                yield 'set filter', f
                for i in range(script['nsubs']):
                    yield 'expose long'

    def frame_generator(self):
        yield 'set filter', 'L'
        yield 'set exposure', self.scripts['frame']['exposure']
        while True:
            yield 'expose short'

    def dark_generator(self):
        yield 'set filter', 'dark'
        yield 'set exposure', self.scripts['dark']['exposure']
        while True:
            yield 'expose long'

    def bias_generator(self):
        yield 'set filter', 'dark'
        yield 'set exposure', self.scripts['bias']['exposure']
        while True:
            yield 'expose bias'

    def flat_generator(self):
        yield 'set filter', self.scripts['flat']['filter'][0]
        yield 'autoflat'
        while True:
            yield 'expose long'

    def reset_generator(self):
        self.generator = getattr(self, '{:}_generator'.format(self.current_script))()




''' Basic class to draw reticle
'''

from kivy.properties import NumericProperty, BooleanProperty
from kivy.lang import Builder
from kivy.uix.widget import Widget

# NB for kivy prior to RC 2 could use :set inner dp(10)
Builder.load_string('''

#:set inner 10

<Reticle>:
    canvas:
        Color: 
            rgba: app.highlight_color if root.show else self.disabled_color
        Line:
            circle: self.x, self.y, inner, 0, 360, 100
        Line:
            circle: self.x, self.y, 3 * inner, 0, 360, 100
        Line:
            circle: self.x, self.y, 6 * inner, 0, 360, 100
        Line:
            points: self.x, self.y - inner, self.x, self.y - 2 * self.radius 
        Line:
            points: self.x, self.y + inner, self.x, self.y + 2 * self.radius 
        Line:
            points: self.x - 2 * self.radius , self.y, self.x - inner, self.y
        Line:
            points: self.x + 2 * self.radius , self.y, self.x + inner, self.y
    disabled_color: [0, 0, 0, 0]

''')

class Reticle(Widget):

    radius = NumericProperty(1)
    show = BooleanProperty(False)



''' Special-purpose widgets for Jocular
'''

import math
from math import radians
from functools import partial

from kivy.app import App
from kivy.core.window import Window
from kivy.clock import Clock
from kivy.uix.anchorlayout import AnchorLayout
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.popup import Popup
from kivy.lang import Builder
from kivy.uix.label import Label
from kivy.properties import (
    NumericProperty,
    ObjectProperty,
    BooleanProperty,
    StringProperty,
    ListProperty,
)
from kivy.uix.button import Button
from kivy.uix.bubble import Bubble, BubbleButton
from kivy.uix.togglebutton import ToggleButton
from kivy.uix.widget import Widget
from kivy.vector import Vector
from kivy.graphics.transformation import Matrix
from kivy.metrics import dp

from jocular.utils import angle360
from jocular.metrics import Metrics

Builder.load_string(
    '''

#:import Clipboard kivy.core.clipboard.Clipboard

<Rotatable>:
    canvas.before:
        PushMatrix
        Translate:
            x: -self.width/2
            y: -self.height/2
        Rotate:
            angle: self.rotangle if self.rotated else 0
            origin: self.center
    canvas.after:
        PopMatrix

<JWidget>:
    # canvas.before:
    #     Color:
    #         rgb: 1, 0, 0
    #         a: .9
    #     Ellipse:
    #         pos: self.pos
    #         size: self.size
    size_hint: None, None
    highlight_color: app.highlight_color
    background_color: .6, 0, .6, 0
    color: app.lowlight_color
    markup: True
    halign: 'center'
    valign: 'middle'
    size: self.texture_size
    font_size: app.ring_font_size
    text_size: None, None  # forces size to be that of text
    padding: 2, 2
    disabled_color: 0, 0, 0, 1

<JFilterButton>:
    canvas.before:
        Color:
            rgb: self.filter_color
            a: .9
        Ellipse:
            pos: self.pos
            size: self.size
    markup: True
    color: (.1, .1, .1, 1)
    background_color: 0, 0, 0, 0
    size: dp(28), dp(28)
    font_size: '16sp'
    size_hint: None, None

<JLever>:
    color: app.lever_color

<JIconButton>:
    text: self.icon

<JToggleButton>:
    color: app.highlight_color if self.state == 'down' else app.lowlight_color

<ELabel>:
    size_hint: 1, None
    markup: True
    text_size: self.size

<ParamValue>:
    size_hint: 1, None
    height: dp(20)
    Button:
        size_hint: None, 1
        markup: True
        halign: 'left'
        text_size: self.size
        color: app.lowlight_color
        font_size: app.info_font_size
        background_color: .6, 0, .6, 0
        width: dp(65)
        padding: dp(3), dp(1)
        text: '[b]{:}[/b]'.format(root.param)
        on_press: root.callback() if root.callback else root.null_action()
    Button:
        size_hint: None, 1
        markup: True
        halign: 'left'
        text_size: self.size
        color: app.lowlight_color
        font_size: app.info_font_size
        background_color: .6, 0, .6, 0
        width: dp(220)
        text: '{:}'.format(root.value)

'''
)


class ParamValue(BoxLayout):

    param = StringProperty('')
    value = StringProperty('')
    callback = ObjectProperty(None)

    def null_action(self, *args):
        pass


class Polar:
    def __init__(self, origin=[0, 0], angle=0, radius=0):
        self.origin = origin
        self.angle = angle
        self.radius = radius

    def to_pos(self):
        ang = self.angle * math.pi / 180
        return (
            self.origin[0] + self.radius * math.cos(ang),
            self.origin[1] + self.radius * math.sin(ang),
        )

    def dist_to(self, pos):
        return Vector(self.to_pos()).distance(pos)


class Rotatable(Widget, Polar):

    rotangle = NumericProperty(0.0)
    rotated = BooleanProperty(True)
    inv = ObjectProperty(Matrix())

    deltas = {
        (1, True): 0,
        (1, False): -90,
        (2, True): -180,
        (2, False): -90,
        (3, True): -180,
        (3, False): 90,
        (4, True): 0,
        (4, False): 90,
    }

    def __init__(self, radial=False, **kwargs):
        super().__init__(**kwargs)
        self.radial = radial
        self.on_angle()

    def on_angle(self, *args):
        q = min(4, (self.angle // 90) + 1)
        self.pos = self.to_pos()
        self.rotangle = self.angle + self.deltas[(q, self.radial)]
        self.generate_inverse()

    def on_width(self, *args):
        self.generate_inverse()

    def on_height(self, *args):
        self.generate_inverse()

    def relocate(self, origin=None, radius=None):
        self.origin = origin
        self.radius = radius
        # self.pos = self.to_pos() # is this causing issues on Windows?
        ang = self.angle * math.pi / 180
        self.x = self.origin[0] + self.radius * math.cos(
            ang
        )  #  is this causing issues on Windows?
        self.y = self.origin[1] + self.radius * math.sin(
            ang
        )  #  is this causing issues on Windows?
        self.generate_inverse()

    def generate_inverse(self, *args):
        # create forward transformation matrix
        # NB this is done on width and height as well as init
        # because when using texture size to compute size
        # correct w/h info is not available on init!

        ang = radians(self.rotangle) if self.rotated else 0
        t = Matrix().translate(-self.width / 2 - self.x, -self.height / 2 - self.y, 0)
        t = Matrix().rotate(ang, 0, 0, 1).multiply(t)
        t = Matrix().translate(self.x, self.y, 0).multiply(t)
        self.inv = t.inverse()

    def collide_point(self, x, y):
        # transform touchpos using the inverse transform of the rotation
        # takes about 150 microseconds

        xt, yt, _ = self.inv.transform_point(x, y, 0)
        return (self.x < xt < (self.x + self.width)) and (
            self.y < yt < (self.y + self.height)
        )


class JWidget(Widget):
    pass


class JRotWidget(JWidget, Rotatable):
    pass


class JLabel(JRotWidget, Label):
    pass


class JToggleButton(ToggleButton, JRotWidget):
    pass


class JButton(Button, JRotWidget):
    pass


class JIconTextButton(JButton, JRotWidget):
    pass


class JFilterButton(Rotatable, Button):
    filter_color = ListProperty([0.2, 0.2, 0.2, 1])


class JIconButton(JButton, JRotWidget):
    icon = StringProperty('')


class JMulti(JButton):
    def __init__(self, values=None, **kwargs):
        super().__init__(**kwargs)
        self.register_event_type('on_press')
        if values is None:
            self.values = []
        else:
            self.values = values
        if self.text not in self.values:
            self.values.append(self.text)  #  default is a single state button

    def on_touch_down(self, touch):
        if super().collide_point(*touch.pos):
            #  find location of self.text in self.values
            if len(self.values) > 1:
                loc = [i for i, x in enumerate(self.values) if x == self.text][0] + 1
                if loc == len(self.values):
                    loc = 0
                self.text = self.values[loc]
            self.dispatch('on_press')
            return True
        return False

    def on_press(self):
        pass


def angle_diff(t1, t2):
    # anticlockwise angle between 2 positions relative to 0 = positive x-axis
    return angle360(math.atan2(t2[1] - t1[1], t2[0] - t1[0]) / (math.pi / 180))


class JLever(JRotWidget, Label):

    value = NumericProperty(0.0)
    disabled = BooleanProperty(False)

    def __init__(
        self, value=0, values=[0, 1], angles=[0, 1], radial=True, springy=None, **kwargs
    ):

        super().__init__(radial=radial, **kwargs)
        self.min_angle, self.max_angle = angles[0], angles[1]
        self.min_value, self.max_value = values[0], values[1]
        self.font_size = '18sp'
        self.springy = springy if springy else None
        self.value = value
        self.selected = False
        self._k = (self.max_angle - self.min_angle) / (self.max_value - self.min_value)
        self.rotangle = self.value_to_angle(self.value)
        self.update(self.rotangle)

    def reset_value(self, v):
        # force update of lever position
        ang = self._k * (v - self.min_value) + self.min_angle
        self.update(ang)

    def value_to_angle(self, v):
        return self._k * (v - self.min_value) + self.min_angle

    def on_touch_down(self, touch):
        if not self.disabled and self.collide_point(*touch.pos):
            self.selected = True
            return True
        return False

        # WAS THIS
        # if self.disabled:
        #     return super().on_touch_down(touch)

        # if self.collide_point(*touch.pos):
        #     self.selected = True
        #     return True
        # return super().on_touch_down(touch)

    def on_touch_move(self, touch):
        if self.selected:
            self.update(angle_diff(self.origin, touch.pos))

    def ang2value(self, angle):
        # x = float(((angle - self.min_angle) / self._k) + self.min_value)
        x = ((angle - self.min_angle) / self._k) + self.min_value
        return min(self.max_value, max(x, self.min_value))

    def on_touch_up(self, touch):
        if self.selected:
            if self.springy is not None:
                self.update(self.value_to_angle(0))
            self.selected = False

    def update(self, angle, update_value=True):
        if update_value:
            mina, maxa = self.min_angle, self.max_angle

            # return if angle outside range
            if (angle < min(mina, maxa)) or (angle > max(mina, maxa)):
                return

            self.value = self.ang2value(angle)

            # convert value to angle in case it hasn't changed
            angle = self.value_to_angle(self.value)

        self.angle = angle
        self.on_angle()


class ELabel(Label):
    pass


class JPopup(Popup):
    def __init__(
        self,
        posn=None,
        message=None,
        actions=None,
        show_for=None,
        cancel_label=None,
        show_title=True,
        **args
    ):

        if 'size' not in args:
            self.size = dp(300), dp(400)

        # user has provided no explicit content, so build it up from message or actions
        if 'content' not in args:
            args['content'] = hl = BoxLayout(
                size_hint=(1, None), height=dp(32), spacing=10
            )

            # use a simple label
            if message is not None:
                hl.add_widget(ELabel(text=message))
                # self.size = dp(300), dp(200)

            elif actions is not None:
                for msg, action in actions.items():
                    b = Button(text=msg)
                    b.bind(on_press=partial(self.do_action, action))
                    hl.add_widget(b)

            self.size = dp(300), dp(150)

            # add cancel if not provided
            if cancel_label is not None:
                hl.add_widget(Button(text=cancel_label, on_press=self.dismiss))

        # now we have content, so place it in an anchor layout (we have to store size first)
        content_width, content_height = args['content'].size
        a = AnchorLayout(anchor_x='center', anchor_y='center')
        a.add_widget(args['content'])
        args['content'] = a

        # generate popup
        super().__init__(
            title_align='center',
            title_size='18sp' if show_title else '0sp',
            separator_height='2dp' if show_title else '0dp',
            size_hint=(None, None),
            auto_dismiss=False,
            separator_color=App.get_running_app().highlight_color,
            **args
        )

        if 'size' not in args:
            if content_width < 200:
                self.width = content_width + dp(200)
            else:
                self.width = content_width + dp(100)
            if show_title:
                self.height = content_height + dp(120)
            else:
                self.height = content_height

        if 'pos_hint' not in args:
            if posn is None:
                self.pos_hint = {'center_x': 0.5, 'center_y': 0.5}
            elif type(posn) == str:
                if posn == 'bottom-left':
                    self.pos_hint = {'x': 0, 'y': 0}
                elif posn == 'bottom-middle':
                    self.pos_hint = {'center_x': 0.5, 'y': 0}
                elif posn == 'top-right':
                    self.pos_hint = {'right': 1, 'top': 1}
                elif posn == 'top-left':
                    self.pos_hint = {'x': 0, 'top': 1}
            else:
                self.pos_hint = {
                    'centre_x': posn[0] / Window.width,
                    'center_y': posn[1] / Window.height,
                }

        if show_for is not None:
            Clock.schedule_once(self.dismiss, show_for)

    def do_action(self, action, *args):
        action()
        self.dismiss()


class JBubble(Widget):
    def __init__(
        self, message=None, actions=None, show_for=2, loc='right', show_cancel=True
    ):
        super().__init__()

        self.hide_bubble(0)

        self.bubble = Bubble(
            size_hint=(None, None),
            orientation='vertical',
            background_color=(0.2, 0.2, 0.2, 1),
            background_image='',
        )
        self.message = message
        self.loc = loc
        self.show_for = show_for

        if message is None:
            if show_cancel and 'Cancel' not in actions:
                actions['Cancel'] = self.hide_bubble
            self.bubble.size = dp(140), dp(25 * len(actions))
            for c, callback in actions.items():
                bb = BubbleButton(text=c)
                self.bubble.add_widget(bb)
                bb.bind(on_press=partial(self.hide_bubble_and_respond, callback))
        else:
            self.bubble.add_widget(BubbleButton(text=message))
            self.bubble.size = dp(10 * len(message)), dp(35)

        cx, cy = Metrics.get('origin')
        if loc == 'right':
            self.bubble.center_y = cy
            self.bubble.x = cx + Metrics.get('radius')
            self.bubble.arrow_pos = 'left_mid'
        elif loc == 'bottom':
            self.bubble.center_x = cx
            self.bubble.y = dp(60)
            self.bubble.arrow_pos = 'bottom_mid'
        elif loc == 'mouse':
            x, y = Window.mouse_pos
            self.bubble.x = x + dp(10)
            self.bubble.center_y = y
            self.bubble.arrow_pos = 'left_mid'
        elif loc == 'center':
            self.bubble.center_x = cx
            self.bubble.center_y = cy

    def open(self):
        if self.message is not None:
            self.clock_event = Clock.schedule_once(self.hide_bubble, self.show_for)
        App.get_running_app().gui.add_widget(self.bubble)

    def hide_bubble_and_respond(self, callback, *args):
        self.hide_bubble(0)
        callback(*args)

    def hide_bubble(self, dt):
        if hasattr(self, 'bubble') and (self.bubble is not None):
            App.get_running_app().gui.remove_widget(self.bubble)


joc_icons = {
    'pause': '3',
    'play': 'n',
    'oculus': 'o',
    'last_sub': 'f',
    'first_sub': 'e',
    'next_sub': '>',
    'prev_sub': '<',
    'camera': 'b',
    'reticle': 'v',
    'fit': 'y',
    'stack': 'E',
    'clear': '0',
    'quit': 'Q',
    'settings': 'x',
    'snapshotter': 'o',
    'prev': 'F',
    'list': 'w',
    'lever': '1',
    'redo': 'h',  # was '4'
    'new': 'i',
    'warn': '!',
    'error': 'W',
    'solve': 'T',
    'slew': 'H',
    'info': '}',
}


def jicon(nm, font_size=None):
    fs = '{:}sp'.format(font_size) if font_size else '16sp'
    icon = joc_icons.get(nm, 'm')
    return "[font=Jocular][size={:}]{:}[/size][/font]".format(fs, icon)



''' Class to construct circular parts of the eyepiece
'''

from kivy.uix.widget import Widget
from kivy.lang import Builder
from kivy.properties import NumericProperty

Builder.load_string('''

<Ring>:
    canvas:
        Color:
            rgba: self.color
            rgba: self.grey, self.grey, self.grey, 1 - app.transparency
        Line:
            circle: self.x, self.y, self.radius, self.start_angle, self.end_angle, self.nsegs
            width: self.thickness
''')

class Ring(Widget):

    thickness = NumericProperty(0)

    def __init__(self, pos=[0, 0], radius=1, thickness=1, grey=.4, 
        start_angle=0, end_angle=360, nsegs=240):
        self.selected = False
        self.pos = pos
        self.radius = radius
        self.thickness = thickness
        self.grey = grey
        self.color = [grey, grey, grey, 1]
        self.start_angle = start_angle
        self.end_angle = end_angle
        self.nsegs = nsegs
        super().__init__()

    def relocate(self, origin=None, radius=None, thickness=None):
        self.pos = origin
        self.radius = radius
        if thickness:
            self.thickness = thickness



''' Various stretch functions. Easy to add more. Room for refinement,
    methinks.
'''

import numpy as np

def stretch(x, method='linear', param=None, NR=0, background=None):

    # if no noise reduction just use stretch alone
    if (NR <= 0) or (background is None):
        return stretch_main(x, method=method, param=param)

    else:
        # get stretched data and lightly suppress low end
        y = stretch_main(x, method=method, param=param)
        hyper_param = 1 - .1 * (NR / 100)
        return y * stretch_main(x, method='hyper', param=hyper_param)

def stretch_main(x, method='linear', param=None):

    if method == 'linear':
        return x

    if method == 'hyper':
        d = .02
        c = d * (1 + d - param)
        return (1 + c) * (x / (x + c))

    if method == 'log':
        c = param * 200
        return np.log(c*x + 1) / np.log(c + 1)

    if method == 'asinh':
        # c = param * 250
        c = param * 2000
        return np.arcsinh(c*x) / np.arcsinh(c + .0000001)

    if method == 'gamma':
        # with noise reduction linear from x=0-a, with slope s
        y = x.copy()
        # g = .5 - .5 * param
        g = .75 - .75 * param
        a0 = .01
        s = g / (a0 * (g - 1) + a0 ** (1 - g))
        d = (1 / (a0 ** g * (g - 1) + 1)) - 1
        y[x < a0] = x[x < a0] * s
        y[x >= a0] =  (1 + d) * (x[x >= a0] ** g) - d
        return y
    
    else:
        return x




'''  Fast gradient and background estimation
'''

import numpy as np
from numpy.polynomial import polynomial
from kivy.logger import Logger
from jocular.utils import percentile_clip

def estimate_background(im):
    # requires about .7ms for Lodestar image on MacBook Pro 2020

     # fit at random pixels and remove outliers
    npts = 500
    r, c = im.shape

    x = np.random.random_integers(0, high=c-1, size=(npts))
    y = np.random.random_integers(0, high=r-1, size=(npts))
    z = np.array([im[j, i] for i, j in zip(x, y)])

    # remove outliers e.g. stars
    zlo, zhi = np.percentile(z, 20), np.percentile(z, 80)
    inds = [i for i, j in enumerate(z) if (j >= zlo) and (j <= zhi)]
    zinds = z[inds]

    return np.mean(zinds), np.std(zinds)
   

def estimate_gradient(im):
    # simple estimation of 2d linear gradient surface ~12ms

    def _polyfit2d(x, y, f, deg):
        deg = np.asarray(deg)
        vander = polynomial.polyvander2d(x, y, deg)
        vander = vander.reshape((-1, vander.shape[-1]))
        f = f.reshape((vander.shape[0],))
        c = np.linalg.lstsq(vander, f, rcond=None)[0]
        return c.reshape(deg+1)

    try:
        # fit at random pixels
        npts = 500
        r, c = im.shape

        x = np.random.random_integers(0, high=c-1, size=(npts))
        y = np.random.random_integers(0, high=r-1, size=(npts))
        z = np.array([im[j, i] for i, j in zip(x, y)])

        zlo, zhi = np.percentile(z, 20), np.percentile(z, 80)
        inds = [i for i, j in enumerate(z) if (j >= zlo) and (j <= zhi)]

        # perform fit
        return polynomial.polyval2d(*np.meshgrid(np.arange(c), np.arange(r)), 
            _polyfit2d(x[inds], y[inds], z[inds], deg=[2, 1]))

    except Exception as e:
        Logger.debug('Gradient: gradient estimation failed; returning zeros ({:})'.format(e))
        return np.zeros(im.shape)


def image_stats(im):

    mean_back, std_back = estimate_background(im)
    imr = im.ravel()

    return {
        'background': mean_back,
        '100 x std. dev.': 100 * std_back,
        'central 75%': percentile_clip(imr, perc=75),
        '99.99 percentile': np.percentile(imr, 99.99),
        '99.9 percentile': np.percentile(imr, 99.9),
        'min': np.min(imr),
        'max': np.max(imr)
        }



''' Astrometry utilities. Will support future extension to charting.
'''

import numpy as np
from scipy.special import cotdg

def to360(v):
    if type(v) == list or type(v) == np.ndarray:
        v[v > 360] = v[v > 360] - 360
        v[v < 0] = v[v < 0] + 360
    else:
        if v > 360:
            v -= 360
        elif v < 0:
            v += 360
    return v

def spherical_distance(ra, dec, ra0, dec0):
    # return distances of points in ra, dec arrays to point (ra0, dec0)
    d1 = np.sin(np.abs(np.radians(dec) - np.radians(dec0)) / 2) ** 2
    d2 = np.cos(np.radians(dec0)) * np.cos(np.radians(dec)) * (np.sin(np.abs(np.radians(ra) - np.radians(ra0)) / 2) ** 2)
    return np.degrees(2 * np.arcsin((d1 + d2) ** .5))

def Eq2Cart(ra, dec, ra0, dec0):
    # conversion from RA/Dec to standard coords; Snyder (1987) ch 22
    # from matlab version; not fully tested

    rads = np.pi / 180.
    dec = rads * dec
    ra_norm = rads * (ra - ra0)

    if dec0 == 90: 
        # north or south pole so use polar gnomic
        x = np.cos(dec) * np.sin(ra_norm)
        y = -cotdg(dec / rads) * np.cos(ra_norm)

    elif dec0 == -90:
        x = -np.cos(dec) * np.sin(ra_norm)
        y = cotdg(dec / rads) * np.cos(ra_norm)
        
    else:
        # oblique gnomic
        dec0 = rads * dec0
        cos0 = np.cos(dec0)
        sin0 = np.sin(dec0)
        cosdec = np.cos(dec)
        sindec = np.sin(dec)
        cos_ra_norm = np.cos(ra_norm)
        den = cos0 * cosdec * cos_ra_norm + sin0 * sindec
        x = (cosdec * np.sin(ra_norm)) / den
        y = (cos0 * sindec - sin0 * cosdec * cos_ra_norm ) / den

    return -x, y  # reverse RA axis

def Cart2Eq(x, y, ra0, dec0):
    # convert from standard coords to RA/Dec (Snyder ch22)
        
    rads = np.pi / 180.

    # unreverse 'ra'
    x = -x
    rho = (x**2 + y**2)**.5
    c = np.arctan(rho)
    cosc, sinc = np.cos(c), np.sin(c)
    
    if dec0 == 90:
        dec = np.arcsin(cosc)
        ra = np.arctan2(x, -y)
        
    elif dec0 == -90:
        dec = np.arcsin(-cosc)
        ra = np.arctan2(x, y)
        
    else:
        cos0, sin0 = np.cos(rads * dec0), np.sin(rads * dec0)
        dec = np.arcsin(cosc * sin0 + ((y * sinc * cos0 / rho)))
        ra = np.arctan2(x * sinc, rho * cosc * cos0 - y * sinc * sin0)
        
    dec /= rads
    ra = ra0 + ra / rads
    dec[rho==0] = dec0
    ra[rho==0] = ra0
    ra[ra > 360] -= 360
    ra[ra < 0] += 360
    return ra, dec


def make_tile(ra0, dec0, fov=3):
    ''' A tile represents all pertinent information about a field
        centred on ra0, dec0 with given fov
    '''
    rads = np.pi / 180.
    fov2 = fov / 2
    polar = ((dec0 + fov2) >= 90) | ((dec0 - fov2) <= -90)
    northern = dec0 >= 0
    if polar:
        min_ra, max_ra = 0, 360
        if northern:
            min_dec, max_dec = 90 - fov2, 90
        else:
            min_dec, max_dec = -90, -90 + fov2
    else:
        ra_width = fov2 / np.cos(dec0 * rads)
        min_ra, max_ra =  to360(ra0 - ra_width), to360(ra0 + ra_width)
        min_dec, max_dec = dec0 - fov2, dec0 + fov2
    return {
        'fov': fov, 'min_ra': min_ra, 'max_ra': max_ra,
        'min_dec': min_dec, 'max_dec': max_dec, 'ra0': ra0, 'dec0': dec0,
        'polar': polar, 'northern': northern
    }

''' to do: combine these functions with those used in Aligner -- not
    clear any more why they are defined here.
'''

def compute_centroids(im, stars):
    # extract info from all candidate stars and return centroids
    r = 8  # centroid patch radius
    patch_size = 2 * r + 1
    grid_x, grid_y = np.meshgrid(np.arange(patch_size), np.arange(patch_size))
    mask = np.sqrt((r - grid_x)**2 + (r - grid_y)**2) <= r
    rows, cols = im.shape
    ptr = 0
    xdata =[]
    ydata =[] 
    mags=[]
    for x, y in stars:
        x, y = int(x), int(y)
        if (x >= r) and (y >= r) and (x < (cols - r)) and (y < (rows - r)):
            imfrag = im[(y - r):(y + r + 1), (x - r):(x + r + 1)].copy()
            cx, cy, mag = compute_centroid(imfrag, mask, grid_x, grid_y)
            if cx > 0:
                xdata.append(cx + x - r)
                ydata.append(cy + y - r)
                mags.append(mag)
                ptr += 1
    return np.array(xdata), np.array(ydata), np.array(mags)

def compute_centroid(imfrag, mask, grid_x, grid_y):
    mean_intensity = np.mean(imfrag[mask])

    # star pixels are in mask and above mean intensity
    candidates = mask & (imfrag > mean_intensity)

    if candidates.any():
        # estimate background and subtract from image
        mean_background = np.mean(imfrag[np.logical_not(candidates)])
        imfrag_sub = candidates * (imfrag - mean_background)       
        wim = np.sum(imfrag_sub)
        wmean = np.mean(imfrag_sub)
        cx = np.sum(grid_x * imfrag_sub) / wim
        cy = np.sum(grid_y * imfrag_sub) / wim
        return cx, cy, wmean

    return -1, -1, -1



''' Miscellany
'''

import os
import numpy as np
from scipy.stats import trimboth
from kivy.logger import Logger

def angle360(angle):
    if angle < 0:
        return angle + 360
    elif angle > 360:
        return angle - 360
    else:
        return angle

def percentile_clip(a, perc=80):
    return np.mean(trimboth(np.sort(a, axis=0), (100 - perc)/100, axis=0), axis=0)    

def unique_member(l):
    # if list has a single unique member, return it, otherwise None
    if len(set(l)) == 1:
        return l[0]
    else:
        return None

def add_if_not_exists(path):
    if not os.path.exists(path):
        try:
            os.mkdir(path)
        except FileExistsError:
            raise FileExistsError

def make_unique_filename(path):
    if not os.path.exists(path):
        return path
    root, name = os.path.split(path) 
    base, ext = os.path.splitext(name)
    cnt = 1
    while os.path.exists(os.path.join(root, '{:}_{:}{:}'.format(base, cnt, ext))):
        cnt += 1
    return os.path.join(os.path.join(root, '{:}_{:}{:}'.format(base, cnt, ext)))

def purify_name(nm):
    # replace any chars that could cause problems with filesystem by spaces
    return nm.replace(':',' ').replace('/',' ').replace('\\',' ').strip()

# consider combining with above; called in ObjectIO
def generate_observation_name(path, prefix=None):

    # use names of form obs 1, obs 2 if prefix is obs
    if (prefix is None) or (prefix == 'light') or (len(prefix) == 0):
        n = 1
        while os.path.exists(os.path.join(path, 'obs {:}'.format(n))):
            n += 1
        return 'obs {:}'.format(n)

    # otherwise using name is possible
    if not os.path.exists(os.path.join(path, prefix)):
        return prefix

    # otherwise add a version number
    n = 1
    while os.path.exists(os.path.join(path, '{:} v{:}'.format(prefix, n))):
        n += 1
    return '{:} v{:}'.format(prefix, n)

def s_to_minsec(s):
    mins, secs = divmod(round(s), 60)
    if mins == 0:
        return '{:}s'.format(secs)
    elif secs == 0:
        return '{:}m'.format(mins)
    else:
        return '{:d}m{:d}s'.format(mins, secs)

def move_to_dir(frompath, topath):
    # move from frompath to topath, making topath dir and creating unique name if necessary

    try:
        toparent, tobase = os.path.split(topath)
        fromparent, filename = os.path.split(frompath)

        # empty parent implies use fromparent
        if not toparent:
            # treat as subdirectory of from
            topath = os.path.join(fromparent, tobase)
        add_if_not_exists(topath)

        # create unique filename
        dest = make_unique_filename(os.path.join(topath, filename))

        # move
        os.rename(frompath, dest)        

    except Exception as e:
        Logger.error('Utils: problem moving {:} to {:} ({:})'.format(frompath, topath, e))
